// INCLUDES---------------------------------------------------------------

#include "resource.h"           // resources

#include "ibm.h"
#define EXEC_TYPES_H
#include "ww.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "ww_strings.h"

#include "bass.h"

#include <stdio.h>              // FILE*
#include <mmsystem.h>           // multimedia timers
#include <time.h>               // for randomization
#include <fcntl.h>              // _O_TEXT
#include <io.h>                 // for _open_osfhandle()
#include <assert.h>

// DEFINES----------------------------------------------------------------

#define SPI_GETKEYBOARDCUES 0x100A
#define SPI_SETKEYBOARDCUES 0x100B

#define CHANNELS     24

#define LANGUAGE_ENG  0
#define LANGUAGE_GER  1
#define LANGUAGE_GRE  2
#define LANGUAGE_ITA  3
#define LANGUAGE_POL  4
#define LANGUAGE_SPA  5
#define LANGUAGES     6

#define INTROSPEED  200 // delay between frames in msec

#define MAKE_ID(a,b,c,d) ((ULONG) (d)<<24 | (ULONG) (c)<<16 | (ULONG) (b)<<8 | (ULONG) (a))
// note that MAKE_ID() is defined differently for IBM-PC because of endianness issues!

#define ID_FORM             MAKE_ID('F', 'O', 'R', 'M')
#define ID_CTLG             MAKE_ID('C', 'T', 'L', 'G')
#define ID_STRS             MAKE_ID('S', 'T', 'R', 'S')

// EXPORTED VARIABLES------------------------------------------------------

EXPORT int                       CatalogPtr, // never used
                                 fontx                        = 8,
                                 joys                         = 0;
EXPORT FLAG                      alt                          = FALSE,
                                 ctrl                         = FALSE,
                                 ignorepaint                  = FALSE,
                                 lmb                          = FALSE,
                                 rmb                          = FALSE,
                                 shift                        = FALSE,
                                 soundfx                      = TRUE,
                                 music                        = TRUE;
EXPORT COLOUR                    currentcolour;
EXPORT TEXT                      currentsentence[SAYLIMIT + 1];
EXPORT HCURSOR                   hArrow,
                                 hPointer,
                                 pointers[10];
EXPORT HMUSIC                    SongHandle[SONGS];
EXPORT ULONG                     DisplayWidth                 = 800,
                                 DisplayHeight                = 600;
EXPORT SBYTE                     resolution                   = 0;
EXPORT const char                g_szClassName[]              = "WormWars";

// IMPORTED VARIABLES------------------------------------------------------

IMPORT FLAG                      aborted,
                                 alwaysrq,
                                 autosave,
                                 clipboarded,
                                 engraved,
                                 isometric,
                                 levels_modified,
                                 scores_modified,
                                 shuffle,
                                 sticky,
                                 superturbo,
                                 tdworms,
                                 turbo;
IMPORT SBYTE                     a,
                                 level,
                                 levels,
                                 players,
                                 reallevel,
                                 sourcelevel;
IMPORT UBYTE                     engraving;
IMPORT UWORD                     field[MAXFIELDX + 1][MAXFIELDY + 1],
                                 gfxfield[MAXFIELDX + 1][MAXFIELDY + 1];
IMPORT SWORD                     fieldx, fieldy,
                                 secondsleft,
                                 secondsperlevel;
IMPORT ULONG                     delay,
                                 difficulty,
                                 quantity[5][3],
                                 r;
IMPORT TEXT                      date[DATELENGTH + 2],
                                 times[TIMELENGTH + 1],
                                 pathname[MAX_PATH + 1],
                                 stattext[7 + 1];
IMPORT STRPTR                    fruitname[LASTFRUIT - FIRSTFRUIT + 1];
IMPORT int                       fex, fey,
                                 introframe,
                                 pseudotop[PSEUDOGADGETS],
                                 showing;
IMPORT struct CreatureInfoStruct creatureinfo[SPECIES + 1];
IMPORT struct HiScoreStruct      hiscore[4][HISCORES + 1];
IMPORT struct ObjectInfoStruct   objectinfo[LASTOBJECT + 1];
IMPORT struct WormStruct         worm[4];

// MODULE VARIABLES--------------------------------------------------------

MODULE HFONT                     hBigFont                     = NULL,
                                 hBoldFont                    = NULL,
                                 hFont                        = NULL,
                                 hGreekFont                   = NULL,
                                 hPolishFont                  = NULL;
MODULE HICON                     BigIconPtr,
                                 icons[12],
                                 SmallIconPtr;
MODULE COLORREF                  statcolour;
MODULE SWORD                     oldsecondsleft               = -1,
                                 xoffset,
                                 yoffset;
MODULE FLAG                      consoleopen                  = FALSE,
                                 freshfire[2]                 = { TRUE, TRUE },
                                 globalabort,
                                 iconified                    = FALSE,
                                 lame                         = FALSE,
                                 musicable                    = FALSE,
                                 paused                       = FALSE,
                                 showpointer                  = TRUE,
                                 WaveOpened[CHANNELS],
                                 WavePrepared[CHANNELS];
MODULE UBYTE                     ConfigBuffer[13];
MODULE SBYTE                     song                         = -1;
MODULE ULONG                     millielapsed;
MODULE UBYTE*                    SoundBuffer[SAMPLES];
MODULE HWAVEOUT                  hWave[CHANNELS];
MODULE WAVEHDR                   WaveHdr[CHANNELS];
MODULE HBRUSH                    hBrush[4],
                                 RedBrush;
MODULE HCURSOR                   hBusy;
MODULE HINSTANCE                 InstancePtr                  = NULL;
MODULE HPEN                      GreyPen                      = NULL;
MODULE HWND                      AnchorWindowPtr              = NULL,
                                 CurrentWindowPtr             = NULL,
                                 MainWindowPtr                = NULL,
                                 TitleWindowPtr               = NULL;
MODULE HMENU                     MenuPtr                      = NULL,
                                 TrayMenuPtr                  = NULL;
MODULE HBITMAP                   BackdropBitmap,
                                 MaskBitmap,
                                 SquaresBitmap,
                                 TDSquaresBitmap;
MODULE HDC                       OnScreenRastPort             = NULL;
MODULE int                       coordfieldx,
                                 coordfieldy,
                                 entirex,
                                 entirey,
                                 highestid                    = 0,
                                 language                     = LANGUAGE_ENG;
MODULE TEXT                      SysInfoName[MAX_PATH + 1];

#define RESOLUTIONS 5
MODULE struct
{   int width, height;
} resolutions[RESOLUTIONS] = {
{  800,  600 },
{ 1024,  600 },
{ 1024,  768 },
{ 1280,  900 },
{ 1280, 1024 },
};

MODULE struct
{
// initialized
    const STRPTR filename;
    const int    menu,
                 codepage;
// uninitialized
          UBYTE* winbuffer;
          ULONG* offset;
          int    size;
} langs[LANGUAGES] =
{ { NULL                      , IDR_MENU_ENG, 1252 },
  { ".\\Catalogs\\german.iff" , IDR_MENU_GER, 1252 },
  { ".\\Catalogs\\greek.iff"  , IDR_MENU_GRE, 1253 },
  { ".\\Catalogs\\italian.iff", IDR_MENU_ITA, 1252 },
  { ".\\Catalogs\\polish.iff" , IDR_MENU_POL, 1250 },
  { ".\\Catalogs\\spanish.iff", IDR_MENU_SPA, 1252 },
};

#define MIB_ICONIFY  3
#define MIB_MANUAL  16
#define MIB_UPDATES 17
#define MIB_ABOUT   18
#define MENUBITMAPS 19 // 0..18
MODULE struct
{   const DWORD   menuitem,
                  resource;
// uninitialized
          HBITMAP bitmap;
} mib[MENUBITMAPS] =
{ { ID_FILE_NEW      , IDB_M_NEW      }, //  0
  { ID_FILE_OPEN     , IDB_M_OPEN     },
  { ID_FILE_SAVE     , IDB_M_SAVE     },
  { ID_FILE_ICONIFY  , IDB_M_ICONIFY  }, //  3 (MIB_ICONIFY)
  { ID_FILE_UNICONIFY, IDB_M_UNICONIFY},
  { ID_FILE_EXITDOS  , IDB_M_EXIT     }, //  5
  { ID_EDIT_CUT      , IDB_M_CUT      },
  { ID_EDIT_COPY     , IDB_M_COPY     },
  { ID_EDIT_PASTE    , IDB_M_PASTE    },
  { ID_EDIT_ERASE    , IDB_M_ERASE    },
  { ID_EDIT_DELETE   , IDB_M_DELETE   }, // 10
  { ID_VIEW_PREV     , IDB_M_PREV     },
  { ID_VIEW_NEXT     , IDB_M_NEXT     },
  { ID_HELP_CREATURES, IDB_M_CREATURES},
  { ID_HELP_OBJECTS  , IDB_M_OBJECTS  },
  { ID_HELP_FRUITS   , IDB_M_FRUITS   }, // 15
  { ID_HELP_MANUAL   , IDB_M_MANUAL   }, // 16 (MIB_MANUAL)
  { ID_HELP_UPDATE   , IDB_M_UPDATES  }, // 17 (MIB_UPDATES)
  { ID_HELP_ABOUT    , IDB_M_ABOUT    }, // 18 (MIB_ABOUT)
};

MODULE struct
{   SCANCODE scancode;
    SBYTE    player, deltax, deltay, special;
    BOOL     down;
} key[NUMKEYS + 1] = {
{SCAN_Q              , 0, -1, -1, MOVE,     FALSE},
{SCAN_W              , 0,  0, -1, MOVE,     FALSE},
{SCAN_E              , 0,  1, -1, MOVE,     FALSE},
{SCAN_A              , 0, -1,  0, MOVE,     FALSE},
{SCAN_S              , 0,  0,  1, MOVE,     FALSE},
{SCAN_D              , 0,  1,  0, MOVE,     FALSE},
{SCAN_Z              , 0, -1,  1, MOVE,     FALSE},
{SCAN_X              , 0,  1,  1, MOVE,     FALSE},
{SCAN_C              , 0,  1,  1, MOVE,     FALSE},
{SCAN_N7             , 1, -1, -1, MOVE,     FALSE},
{SCAN_N8             , 1,  0, -1, MOVE,     FALSE},
{SCAN_N9             , 1,  1, -1, MOVE,     FALSE},
{SCAN_N4             , 1, -1,  0, MOVE,     FALSE},
{SCAN_N5             , 1,  0,  1, MOVE,     FALSE},
{SCAN_N6             , 1,  1,  0, MOVE,     FALSE},
{SCAN_N1             , 1, -1,  1, MOVE,     FALSE},
{SCAN_N2             , 1,  0,  1, MOVE,     FALSE},
{SCAN_N3             , 1,  1,  1, MOVE,     FALSE},
{SCAN_UP             , 1,  0, -1, ONEHUMAN, FALSE},
{SCAN_DOWN           , 1,  0,  1, ONEHUMAN, FALSE},
{SCAN_RIGHT          , 1,  1,  0, ONEHUMAN, FALSE},
{SCAN_LEFT           , 1, -1,  0, ONEHUMAN, FALSE},
{SCAN_SPACEBAR       , 0,  0,  0, AMMO,     FALSE},
{SCAN_V              , 0,  0,  0, AMMO,     FALSE},
{SCAN_N0             , 1,  0,  0, AMMO,     FALSE},
{SCAN_ENTER          , 1,  0,  0, AMMO,     FALSE},
{SCAN_HELP           , 0,  0,  0, TRAINER,  FALSE},
{SCAN_NUMERICSLASH   , 0,  0,  0, TRAINER,  FALSE},
{SCAN_NUMERICASTERISK, 0,  0,  0, TRAINER,  FALSE},
{SCAN_NUMERICPLUS    , 0,  0,  0, TRAINER,  FALSE},
{SCAN_NUMERICMINUS   , 0,  0,  0, TRAINER,  FALSE}
};

MODULE struct
{   // initialized
    STRPTR filename;
    // uninitialized
    int    hertz,
           length;
} samp[SAMPLES] =
{ {      "fx\\dog.wav"}, // FXBORN_DOG
  {     "fx\\rain.wav"}, // FXBORN_RAIN
  {"fx\\lightning.wav"}, // FXGET_LIGHTNING
  {  "fx\\missile.wav"}, // FXBORN_MISSILE
  {"fx\\protector.wav"}, // FXBORN_PROTECTOR
  {    "fx\\death.wav"}, // FXDEATH_WORM
  {  "fx\\enclose.wav"}, // FXDO_ENCLOSE
  { "fx\\fragment.wav"}, // FXBORN_FRAGMENT
  {     "fx\\ammo.wav"}, // FXGET_AMMO
  {  "fx\\cyclone.wav"}, // FXGET_CYCLONE
  {  "fx\\getrain.wav"}, // FXGET_RAIN
  {   "fx\\grower.wav"}, // FXGET_GROWER
  {   "fx\\object.wav"}, // FXGET_OBJECT
  {  "fx\\powerup.wav"}, // FXGET_POWERUP
  {    "fx\\grave.wav"}, // FXGET_SKULL
  {   "fx\\amigan.wav"}, // FXHELP
  {    "fx\\shoot.wav"}, // FXUSE_AMMO
  {   "fx\\armour.wav"}, // FXUSE_ARMOUR
  {     "fx\\bomb.wav"}, // FXUSE_BOMB
  { "fx\\teleport.wav"}, // FXUSE_TELEPORT
  {     "fx\\bird.wav"}, // FXBORN_BIRD
  {    "fx\\green.wav"}, // FXPAIN
  {      "fx\\red.wav"},
  {     "fx\\blue.wav"},
  {   "fx\\yellow.wav"},
  { "fx\\applause.wav"}, // FXAPPLAUSE
  { "fx\\gameover.wav"}, // FXGAMEOVER
  {    "fx\\click.wav"}, // FXCLICK
  {    "fx\\siren.wav"}, // FXSIREN
  {     "fx\\ding.wav"}, // FXDING
  {     "fx\\bull.wav"}, // FXBORN_BULL
  {    "fx\\snake.wav"}, // FXBORN_SNAKE
  {    "fx\\rhino.wav"}, // FXBORN_RHINO
  {"fx\\butterfly.wav"}, // FXGET_BUTTERFLY
  { "fx\\elephant.wav"}, // FXBORN_ELEPHANT
  {    "fx\\horse.wav"}, // FXBORN_HORSE
  {   "fx\\monkey.wav"}, // FXBORN_MONKEY
};

MODULE const ULONG BGRColour[4] = { 0x0066CC66, // green
                                    0x008888FF, // red
                                    0x00FF9999, // blue
                                    0x0066CCCC  // yellow
                                  };

// MODULE FUNCTIONS-------------------------------------------------------

MODULE void soundicons(void);
MODULE void load_music(void);
MODULE void load_samples(void);
MODULE ULONG getsize(STRPTR filename);
MODULE void redrawscreen(void);
MODULE void next(HWND hwnd);
MODULE void calcsize(void);
MODULE void loadconfig(void);
MODULE void saveconfig(void);
MODULE void killfocus(HWND hwnd, int focussed);
MODULE void renderhiscores(HWND hwnd);
MODULE void do_pause(void);
MODULE void load_catalog(void);
MODULE void parse_catalog(void);
MODULE void retranslate(void);
MODULE void freelanguage(void);
MODULE void titlescreen(void);
MODULE void uniconify(FLAG exiting);
MODULE void traymenu(void);
MODULE void repaint_intro(void);
MODULE void setdlgtext(HWND hwnd, int gid, int msg, STRPTR defaultmsg);
MODULE void le_handlectrlkybd(SCANCODE scancode);
MODULE FLAG firebutton(void);
MODULE void setcodepage(void);
MODULE void changelanguage(void);
MODULE void confinepointer(void);
MODULE void drawitem(LPDRAWITEMSTRUCT lpDIS);
MODULE void do_systemmenu(HWND hwnd);

MODULE BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
MODULE BOOL CALLBACK TitleDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam);
MODULE LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

// CODE--------------------------------------------------------------------

EXPORT int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{   TEXT          errormessage[SAYLIMIT + 1];
    JOYINFOEX     joyInfoEx;
    int           i, j;
    WNDCLASSEX    wc;
    LANGID        langid;
    BOOL          b,
                  pvParam;
    DWORD         processid;
    OSVERSIONINFO Vers;
    WSADATA       wsaData;

    // Start of program.

    for (i = 0; i < SAMPLES; i++)
    {   SoundBuffer[i] = NULL;
    }
    for (i = 0; i < MENUBITMAPS; i++)
    {   mib[i].bitmap  = NULL;
    }

    InstancePtr = hInstance;
    srand((unsigned) time(NULL));
    processid = GetCurrentProcessId();

    Vers.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx((LPOSVERSIONINFO) &Vers);
    if (Vers.dwPlatformId == VER_PLATFORM_WIN32_NT)
    {   // OC; printf("Lame OS detected!\n");
        lame = TRUE;
        /* if (Vers.dwMajorVersion >= 6)
        {   // OC; printf("Vista/Windows 7/8/8.1/10 detected!\n");
            win7 = TRUE;
        } */
    } elif (Vers.dwPlatformId == VER_PLATFORM_WIN32s)
    {   ; // OC; printf("Please upgrade to Windows 9x!\n");
    } elif (Vers.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
    {   ; // OC; printf("Windows 9x detected.\n");
    } else
    {   // OC; printf("Unknown OS!\n");
        lame = TRUE;
    }

     BackdropBitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_BKGRND   ));
         MaskBitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_TDMASK   ));
      SquaresBitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_SQUARES  ));
    TDSquaresBitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(IDB_TDSQUARES));

    hArrow = LoadCursor(NULL, IDC_ARROW);
    hBusy  = LoadCursor(NULL, IDC_WAIT);

    /* Note that if the font size is greater than 12 points this will overlap
    with the icon()s. */

    hFont = CreateFont
    (   12,                     // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_MEDIUM,              // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        ANSI_CHARSET,           // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "Arial"                 // LPCTSTR lpszFacename
    );
    hBigFont = CreateFont
    (   18,                     // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_EXTRABOLD,           // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        ANSI_CHARSET,           // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "Arial"                 // LPCTSTR lpszFacename
    );
    hBoldFont = CreateFont
    (   8,                      // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_EXTRABOLD,           // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        ANSI_CHARSET,           // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_DECORATIVE,          // BYTE nPitchAndFamily
        "MS Sans Serif"         // LPCTSTR lpszFacename
    );
    hGreekFont = CreateFont
    (   14,                     // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_MEDIUM,              // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        GREEK_CHARSET,          // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_SWISS,               // BYTE nPitchAndFamily
        lame ? "MS Sans Serif" : "Comic Sans MS (Greek)" // LPCTSTR lpszFacename
    );
    hPolishFont = CreateFont
    (   14,                     // int nHeight
        0,                      // int nWidth
        0,                      // int nEscapement
        0,                      // int nOrientation
        FW_MEDIUM,              // int nWeight
        FALSE,                  // BYTE bItalic
        FALSE,                  // BYTE bUnderline
        0,                      // BYTE cStrikeOut
        EASTEUROPE_CHARSET,     // BYTE nCharSet
        OUT_DEFAULT_PRECIS,     // BYTE nOutPrecision
        CLIP_CHARACTER_PRECIS,  // BYTE nClipPrecision
        DEFAULT_QUALITY,        // BYTE nQuality
        FF_SWISS,               // BYTE nPitchAndFamily
        lame ? "MS Sans Serif" : "Comic Sans MS (Polish)" // LPCTSTR lpszFacename
    );

    // we could DestroyIcon() these at exit but Windows does that automatically for us anyway
    SmallIconPtr = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_WORMWARS     ), IMAGE_ICON  , 16, 16, 0);
    BigIconPtr   = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_WORMWARS     ), IMAGE_ICON  , 32, 32, 0);
    icons[ 0]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_LOAD         ), IMAGE_ICON  , 16, 16, 0);
    icons[ 1]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_SAVE         ), IMAGE_ICON  , 16, 16, 0);
    icons[ 2]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_MANUAL       ), IMAGE_ICON  , 16, 16, 0);
    icons[ 3]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_ABOUT        ), IMAGE_ICON  , 16, 16, 0);
    icons[ 4]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_CREATURES    ), IMAGE_ICON  , 16, 16, 0);
    icons[ 5]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_OBJECTS      ), IMAGE_ICON  , 16, 16, 0);
    icons[ 6]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_FRUITS       ), IMAGE_ICON  , 16, 16, 0);
    icons[ 7]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_STARTGAME    ), IMAGE_ICON  , 16, 16, 0);
    icons[ 8]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_LEVELEDIT    ), IMAGE_ICON  , 16, 16, 0);
    icons[ 9]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_UPDATE       ), IMAGE_ICON  , 16, 16, 0);
    icons[10]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_CLEARHISCORES), IMAGE_ICON  , 16, 16, 0);
    icons[11]    = LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_OK           ), IMAGE_ICON  , 16, 16, 0);

    // we could DestroyCursor() these at exit but Windows does that automatically for us anyway
    pointers[0]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_EMPTY        ), IMAGE_CURSOR, 12, 12, 0);
    pointers[1]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_SILVER       ), IMAGE_CURSOR, 12, 12, 0);
    pointers[2]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_GOLD         ), IMAGE_CURSOR, 12, 12, 0);
    pointers[3]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_DYNAMITE     ), IMAGE_CURSOR, 12, 12, 0);
    pointers[4]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_WOOD         ), IMAGE_CURSOR, 12, 12, 0);
    pointers[5]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_STONE        ), IMAGE_CURSOR, 12, 12, 0);
    pointers[6]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_METAL        ), IMAGE_CURSOR, 12, 12, 0);
    pointers[7]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_FROST        ), IMAGE_CURSOR, 12, 12, 0);
    pointers[8]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_UPARROW      ), IMAGE_CURSOR, 12, 12, 0);
    pointers[9]  = LoadImage(InstancePtr, MAKEINTRESOURCE(IDP_DNARROW      ), IMAGE_CURSOR, 12, 12, 0);

    GreyPen  = CreatePen(PS_SOLID, 0, 0x00808080);
    RedBrush = CreateSolidBrush(RGB(255, 128, 128)); // pens, brushes, etc. are freed at exit

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = InstancePtr;
    wc.hIcon         = BigIconPtr;
    wc.hIconSm       = SmallIconPtr;
    wc.hCursor       =
    hPointer         = hArrow;
    wc.hbrBackground = (HBRUSH) (COLOR_GRAYTEXT + 1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_szClassName;
    if (!RegisterClassEx(&wc))
    {   MessageBox(NULL, "Window registration failed!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

    j = joyGetNumDevs();
    for (i = 0; i < j; i++)
    {   /* DISCARD joyGetDevCaps(i, &JoyCaps, sizeof(JOYCAPS));
        printf("Joystick %d has %d buttons.\n", i, JoyCaps.wNumButtons); */

        joyInfoEx.dwSize  = sizeof(joyInfoEx);
        joyInfoEx.dwFlags = JOY_RETURNALL;
        if (joyGetPosEx(i, &joyInfoEx) == JOYERR_NOERROR)
        {   joys++;
    }   }

    enginesetup();
    langid = GetUserDefaultLangID() & 0xFF;
    switch (langid)
    {
    case  0x07: language = LANGUAGE_GER;
    acase 0x08: language = LANGUAGE_GRE;
    acase 0x0A: language = LANGUAGE_SPA;
    acase 0x10: language = LANGUAGE_ITA;
    acase 0x15: language = LANGUAGE_POL;
    adefault:   language = LANGUAGE_ENG;
    }

    entirex = GetSystemMetrics(SM_CXSCREEN);
    entirey = GetSystemMetrics(SM_CYSCREEN);
    resolution = 0;
    for (i = RESOLUTIONS - 1; i >= 1; i--)
    {   if
        (   GetSystemMetrics(SM_CXSCREEN) >= resolutions[i].width
         && GetSystemMetrics(SM_CYSCREEN) >= resolutions[i].height
        )
        {   resolution = i;
            break;
    }   }

    loadconfig();
    load_catalog();
    enginesetup2();

    DISCARD WSAStartup(MAKEWORD(2,2), &wsaData); // initialize Winsock

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == TRUE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, FALSE, 0, SPIF_SENDWININICHANGE);
    }   }

    if (loadfields(pathname))
    {   newfields();
    }

    if (!(AnchorWindowPtr = CreateWindowEx(
        0,
        g_szClassName,
        NULL,
        0,
        0, 0,
        1, 1,
        NULL,
        NULL,
        InstancePtr, NULL
    )))
    {   MessageBox(NULL, "Window creation failed!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
        cleanexit(EXIT_FAILURE);
    }

    for (i = 0; i < MENUBITMAPS; i++)
    {   mib[i].bitmap = LoadBitmap(InstancePtr, MAKEINTRESOURCE(mib[i].resource));
    }
    SystemParametersInfo(SPI_GETKEYBOARDCUES, 0, &b, 0);
    if (!b)
    {   b = TRUE;
        SystemParametersInfo(SPI_SETKEYBOARDCUES, 0, &b, 0);
    }

    for (i = 0; i <= 3; i++)
    {   hBrush[i] = CreateSolidBrush(BGRColour[i]); // these are freed automatically at exit
    }

    load_samples();
    load_music();
    titlescreen();
}

// SUPPORT FUNCTIONS------------------------------------------------------

EXPORT void cleanexit(SLONG rc)
{   int  i;
    BOOL pvParam;

    if (song >= 0 && song < SONGS)
    {   BASS_ChannelStop(SongHandle[song]);
    }
    BASS_Stop();
    for (i = 0; i < SONGS; i++)
    {   if (SongHandle[song])
        {   BASS_MusicFree(SongHandle[song]);
    }   }
    BASS_Free(); // reverse of BASS_Init();

    if (iconified)
    {   uniconify(TRUE);
    }
    if (MainWindowPtr)
    {   DestroyWindow(MainWindowPtr);
        MainWindowPtr = NULL;
    }

    for (i = 0; i < MENUBITMAPS; i++)
    {   if (mib[i].bitmap)
        {   DeleteObject(mib[i].bitmap);
            mib[i].bitmap = NULL;
    }   }

    DISCARD WSACleanup();

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == FALSE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, TRUE, 0, SPIF_SENDWININICHANGE);
    }   }

    if (TDSquaresBitmap) { DeleteObject(TDSquaresBitmap); TDSquaresBitmap = NULL; }
    if (  SquaresBitmap) { DeleteObject(  SquaresBitmap);   SquaresBitmap = NULL; }
    if (     MaskBitmap) { DeleteObject(     MaskBitmap);      MaskBitmap = NULL; }
    if ( BackdropBitmap) { DeleteObject( BackdropBitmap);  BackdropBitmap = NULL; }

    for (i = 0; i < CHANNELS; i++)
    {   if (WavePrepared[i])
        {   DISCARD waveOutReset(hWave[i]);
            DISCARD waveOutUnprepareHeader(hWave[i], &WaveHdr[i], sizeof(WAVEHDR));
            WavePrepared[i] = FALSE;
        }
        if (WaveOpened[i])
        {   DISCARD waveOutClose(hWave[i]);
            WaveOpened[i] = FALSE;
    }   }
    for (i = 0; i < SAMPLES; i++)
    {   if (SoundBuffer[i])
        {   free(SoundBuffer[i]);
    }   }

    freelanguage();

    DeleteObject(hBigFont);
    DeleteObject(hBoldFont);
    DeleteObject(hFont);
    DeleteObject(hGreekFont);
    DeleteObject(hPolishFont);

    if (rc == EXIT_SUCCESS)
    {   saveconfig();
        // savefields(pathname);
    }

    exit(rc);
}

EXPORT void clearjoystick(void)              { ; }
EXPORT void freefx(void)                     { ; }
EXPORT void renderhiscores_amiga(void)       { ; }
EXPORT void sendreadrequest(void)            { ; }
EXPORT void setpointer(SBYTE brush)          { ; }
EXPORT void ReadAdapterJoystick(UBYTE unit)  { ; }
EXPORT void ReadGamepads(void)               { ; }
EXPORT void ReadStandardJoystick(ULONG unit) { ; }

MODULE FLAG firebutton(void)
{   JOYINFOEX joyInfoEx;

    if (joyGetPosEx(0, &joyInfoEx) == JOYERR_NOERROR)
    {   if (joyInfoEx.dwButtons)
        {   if (freshfire[0])
            {   freshfire[0] = FALSE;
                return TRUE;
        }   }
        else
        {   freshfire[0] = TRUE;
    }   }
    return FALSE;
}

EXPORT void readibmgameport(int unit)
{   TRANSIENT int       deltax, deltay;
    TRANSIENT JOYINFOEX joyInfoEx;
    PERSIST   int       olddeltax[2] = { 0,    0 },
                        olddeltay[2] = { 0,    0 };

    if (worm[2 + unit].control != JOYSTICK || !worm[2 + unit].lives)
    {   return;
    }

    ZeroMemory(&joyInfoEx, sizeof(joyInfoEx));
    joyInfoEx.dwSize  = sizeof(joyInfoEx);
    joyInfoEx.dwFlags = JOY_RETURNALL;

    if (joyGetPosEx(unit, &joyInfoEx) == JOYERR_NOERROR)
    {   if (joyInfoEx.dwButtons)
        {   if (freshfire[unit])
            {   wormqueue((SBYTE) (2 + unit), 0, 0);
                freshfire[unit] = FALSE;
                return;
        }   }
        else
        {   freshfire[unit] = TRUE;
        }

        if     ((joyInfoEx.dwXpos >> 8) <  48)
        {   deltax = -1;
        } elif ((joyInfoEx.dwXpos >> 8) > 208)
        {   deltax =  1;
        } else
        {   deltax =  0;
        }

        if     ((joyInfoEx.dwYpos >> 8) <  48)
        {   deltay = -1;
        } elif ((joyInfoEx.dwYpos >> 8) > 208)
        {   deltay =  1;
        } else
        {   deltay =  0;
        }

        if ((deltax || deltay) && (deltax != olddeltax[unit] || deltay != olddeltay[unit]))
        {   wormqueue((SBYTE) (2 + unit), (SBYTE) deltax, (SBYTE) deltay);
        }
        olddeltax[unit] = deltax;
        olddeltay[unit] = deltay;
}   }

EXPORT void gameinput(void)
{   MSG Message;

    while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
    {   TranslateMessage(&Message);
        DispatchMessage(&Message);
}   }

/* NAME         anykey -- wait for a user press
SYNOPSIS        anykey(FLAG, FLAG);
FUNCTION        Brings up a requester and waits for it to be closed by user.
INPUTS          timeout - whether to timeout automatically after a few
                seconds
                allowrmb - whether to reenable menus at end
IMPLEMENTATION  This function takes control of message processing (see
                "Inside Visual C++", p. 175-176). It doesn't return
                until it receives a WM_KEYDOWN or similar message.
NOTES           This IBM-PC version never times out.
RESULTS         FALSE if Escape was pressed, otherwise TRUE.
FILE            system.cpp */

EXPORT FLAG anykey(FLAG timeout, FLAG allowrmb)
{   return busyloop(TRUE);
}

EXPORT void celebrate(void)
{   int bestscore = -1,
        player,
        winner    =  0,
        x, y;

    for (player = 0; player <= 3; player++)
    {   if (worm[player].control != NONE)
        {   if (worm[player].score >= worm[player].hiscore)
            {   worm[player].hiscore = worm[player].score;
            }
            if (worm[player].score > bestscore)
            {   bestscore = worm[player].score;
                winner = player;
    }   }   }
    say(LLL(MSG_CONGRATULATIONS, "*** CONGRATULATIONS! ***"), worm[winner].colour);
    for (x = 0; x <= fieldx; x++)
    {   for (y = 0; y <= fieldy; y++)
        {   draw(x, y, (UWORD) (FIRSTHEAD + winner));
    }   }
    a = GAMEOVER;
#ifndef TESTING
    waitasec();
    busyloop(TRUE);
#endif
}

EXPORT void clearscreen(void)
{   SWORD   x = 0,
            y;
    HBITMAP OldBitmapPtr;
    HDC     CompatibleRastPort;

    OnScreenRastPort = GetDC(MainWindowPtr);
    CompatibleRastPort = CreateCompatibleDC(OnScreenRastPort);
    OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, BackdropBitmap);
    while (x < DisplayWidth)
    {   y = 0;
        while (y < DisplayHeight)
        {   BitBlt
            (   OnScreenRastPort,
                x,
                y,
                58,
                58,
                CompatibleRastPort,
                0,
                0,
                SRCCOPY
            );
            y += 58;
        }
        x += 58;
    }
    SelectObject(CompatibleRastPort, OldBitmapPtr);
    DeleteDC(CompatibleRastPort);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;

    soundicons();
    dosidebar();
}

EXPORT void draw(int x, int y, UWORD image)
{   HBITMAP OldBitmapPtr;
    HDC     CompatibleRastPort;
    int     mask,
            yy, yyy;
    FLAG    masking;

    if (!MainWindowPtr || superturbo)
    {   return; // this actually does happen at startup
    }

    if (image == EMPTY && engraved)
    {   image = FIRSTPATTERN + engraving;
    }

    // When we call this function, Windows sends a WM_PAINT to our window,
    // if we have never painted this window before.

    OnScreenRastPort = GetDC(MainWindowPtr);
    CompatibleRastPort = CreateCompatibleDC(OnScreenRastPort);

    if (tdworms && image >= HEADSOFFSET)
    {   masking = TRUE;
        image -= HEADSOFFSET;
        if (image < 64) // head
        {   mask = 32 + (image % 8);
        } else // tail
        {   mask = (image - 64) % 29;
    }   }
    else
    {   masking = FALSE;
    }

 /* if (tdworms && image >= FIRSTPAIN && image <= LASTPAIN)
    {   if (x > 0         ) draw(x - 1, y    , gfxfield[x - 1][y    ]);
        if (y > 0         ) draw(x    , y - 1, gfxfield[x    ][y - 1]);
        if (x > 0 && y > 0) draw(x - 1, y - 1, gfxfield[x - 1][y - 1]);
    } */

    if (isometric && x >= 0 && y >= 0)
    {   if (masking)
        {   for (yy = (y <= 0 ? 4 : 0); yy < SQUAREY + 4; yy += 2) // we should probably unroll this loop for speed
            {   OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, MaskBitmap);
                if (x == 0)
                {   BitBlt
                    (   OnScreenRastPort,
                        (x * SQUAREX) + STARTXPIXEL + 2 + ((fieldy - 1 - y) * SQUAREY / 2) + SQUAREX - (yy / 2),
                        (y * SQUAREY) + STARTYPIXEL - 4 + yy,
                        SQUAREX,
                        2,
                        CompatibleRastPort,
                        (mask % 16) * (SQUAREX + 4) + 4,
                        (mask / 16) * (SQUAREY + 4) + yy,
                        SRCAND
                    );
                    SelectObject(CompatibleRastPort, TDSquaresBitmap);
                    BitBlt
                    (   OnScreenRastPort,
                        (x * SQUAREX) + STARTXPIXEL + 2 + ((fieldy - 1 - y) * SQUAREY / 2) + SQUAREX - (yy / 2),
                        (y * SQUAREY) + STARTYPIXEL - 4 + yy,
                        SQUAREX,
                        2,
                        CompatibleRastPort,
                        (image % 16) * (SQUAREX + 4) + 4,
                        (image / 16) * (SQUAREY + 4) + yy,
                        SRCPAINT
                    );
                } else
                {   BitBlt
                    (   OnScreenRastPort,
                        (x * SQUAREX) + STARTXPIXEL - 2 + ((fieldy - 1 - y) * SQUAREY / 2) + SQUAREX - (yy / 2),
                        (y * SQUAREY) + STARTYPIXEL - 4 + yy,
                        SQUAREX + 4,
                        2,
                        CompatibleRastPort,
                        (mask % 16) * (SQUAREX + 4),
                        (mask / 16) * (SQUAREY + 4) + yy,
                        SRCAND
                    );
                    SelectObject(CompatibleRastPort, TDSquaresBitmap);
                    BitBlt
                    (   OnScreenRastPort,
                        (x * SQUAREX) + STARTXPIXEL - 2 + ((fieldy - 1 - y) * SQUAREY / 2) + SQUAREX - (yy / 2),
                        (y * SQUAREY) + STARTYPIXEL - 4 + yy,
                        SQUAREX + 4,
                        2,
                        CompatibleRastPort,
                        (image % 16) * (SQUAREX + 4),
                        (image / 16) * (SQUAREY + 4) + yy,
                        SRCPAINT
                    );
        }   }   }
        else
        {   OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, SquaresBitmap);
            for (yy = 0; yy < SQUAREY; yy += 2) // we should probably unroll this loop for speed
            {   BitBlt
                (   OnScreenRastPort,
                    (x * SQUAREX) + STARTXPIXEL + ((fieldy - 1 - y) * SQUAREY / 2) + SQUAREX - (yy / 2),
                    (y * SQUAREY) + STARTYPIXEL + yy,
                    SQUAREX,
                    2,
                    CompatibleRastPort,
                    (image % 16) * SQUAREX,
                    ((image / 16) * SQUAREY) + yy,
                    SRCCOPY
                );
    }   }   }
    else
    {   if (masking)
        {   OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, MaskBitmap);
            BitBlt
            (   OnScreenRastPort,
                (x * SQUAREX) + STARTXPIXEL - (x <= 0 ? 0 : 4),
                (y * SQUAREY) + STARTYPIXEL - (y <= 0 ? 0 : 4),
                SQUAREX + (x <= 0 ? 0 : 4),
                SQUAREY + (y <= 0 ? 0 : 4),
                CompatibleRastPort,
                ((mask % 16) * (SQUAREX + 4)) + (x <= 0 ? 4 : 0),
                ((mask / 16) * (SQUAREY + 4)) + (y <= 0 ? 4 : 0),
                SRCAND
            );
            SelectObject(CompatibleRastPort, TDSquaresBitmap);
            BitBlt
            (   OnScreenRastPort,
                (x * SQUAREX) + STARTXPIXEL - (x <= 0 ? 0 : 4),
                (y * SQUAREY) + STARTYPIXEL - (y <= 0 ? 0 : 4),
                SQUAREX + (x <= 0 ? 0 : 4),
                SQUAREY + (y <= 0 ? 0 : 4),
                CompatibleRastPort,
                ((image % 16) * (SQUAREX + 4)) + (x <= 0 ? 4 : 0),
                ((image / 16) * (SQUAREY + 4)) + (y <= 0 ? 4 : 0),
                SRCPAINT
            );
        } else
        {   OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, SquaresBitmap);
            BitBlt
            (   OnScreenRastPort,
                (x * SQUAREX) + STARTXPIXEL,
                (y * SQUAREY) + STARTYPIXEL,
                SQUAREX,
                SQUAREY,
                CompatibleRastPort,
                (image % 16) * SQUAREX,
                (image / 16) * SQUAREY,
                SRCCOPY
            );
    }   }

    SelectObject(CompatibleRastPort, OldBitmapPtr);
    DeleteDC(CompatibleRastPort);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void le_draw(int x, int y, UWORD image)
{   HBITMAP OldBitmapPtr;
    HDC     CompatibleRastPort;

    if (!MainWindowPtr)
    {   return; // this actually does happen at startup
    }

    OnScreenRastPort = GetDC(MainWindowPtr);
    CompatibleRastPort = CreateCompatibleDC(OnScreenRastPort);
    OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, SquaresBitmap);
    BitBlt
    (   OnScreenRastPort,
        (x * SQUAREX) + STARTXPIXEL,
        (y * SQUAREY) + STARTYPIXEL,
        SQUAREX,
        SQUAREY,
        CompatibleRastPort,
        (image % 16) * SQUAREX,
        (image / 16) * SQUAREY,
        SRCCOPY
    );
    SelectObject(CompatibleRastPort, OldBitmapPtr);
    DeleteDC(CompatibleRastPort);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void handleinput(SCANCODE code)
{   SBYTE i,
          keyplayer;
    FLAG  allowed = TRUE;

    for (i = 0; i <= NUMKEYS; i++)
    {   key[i].down = FALSE;
    }

    // OC; printf("%d\n", code);
    switch (code)
    {
    case SCAN_LCTRL:
    case SCAN_RCTRL:
        ctrl = TRUE;
    acase SCAN_LSHIFT:
    case SCAN_RSHIFT:
        shift = TRUE;
    acase SCAN_F:
    case SCAN_M:
        toggle((SBYTE) code);
    acase SCAN_T:
        for (i = 0; i <= 3; i++)
        {   if
            (   worm[i].control != IBMPC
             && worm[i].control != NONE
             && worm[i].lives
            )
            {   allowed = FALSE;
                break;
        }   }
        if (allowed)
        {   if (turbo)
            {   turbo = FALSE;
                draw(CLOCKICONX, ICONY, BLACKENED);
            } else
            {   turbo = TRUE;
                draw(CLOCKICONX, ICONY, CLOCK);
        }   }
    acase SCAN_B:
        for (i = 0; i <= 3; i++)
        {   if
            (   worm[i].control != IBMPC
             && worm[i].control != NONE
             && worm[i].lives
            )
            {   allowed = FALSE;
                break;
        }   }
        if (allowed)
        {   if (turbo)
            {   turbo = superturbo = FALSE;
                draw(CLOCKICONX, ICONY, BLACKENED);
            } else
            {   draw(CLOCKICONX, ICONY, CLOCK);
                // order-dependent!
                turbo = superturbo = TRUE;
        }   }
    acase SCAN_P:
    case SCAN_PAUSE:
        do_pause();
    acase SCAN_ESCAPE:
        aborted = TRUE;
        if (level)
        {   for (i = 0; i <= 3; i++)
            {   if (worm[i].lives)
                {   worm[i].levelreached = level;
        }   }   }
        else
        {   for (i = 0; i <= 3; i++)
            {   if (worm[i].lives)
                {   worm[i].levelreached = reallevel;
        }   }   }
        worm[0].lives = worm[1].lives = worm[2].lives = worm[3].lives = 0;
    acase SCAN_F12:
        if (ctrl)
        {   iconify();
        }
    adefault:
        for (i = 0; i <= NUMKEYS; i++)
        {   if (code == key[i].scancode)
            {   key[i].down = TRUE;
                break;
    }   }   }

    for (i = 0; i <= NUMKEYS; i++)
    {   if (key[i].down)
        {   if (key[i].special == ONEHUMAN)
            {   if (worm[0].control == KEYBOARD && worm[1].control != KEYBOARD)
                {   wormqueue(0, key[i].deltax, key[i].deltay);
                } elif (worm[1].control == KEYBOARD)
                {   wormqueue(1, key[i].deltax, key[i].deltay);
            }   }
            elif (key[i].special == MOVE || key[i].special == AMMO)
            {   if (worm[key[i].player].control == KEYBOARD)
                {   keyplayer = key[i].player;
                } elif (key[i].player == 1 && worm[0].control == KEYBOARD && worm[1].control != KEYBOARD)
                {   keyplayer = 0;
                } elif (key[i].player == 0 && worm[1].control == KEYBOARD && worm[0].control != KEYBOARD)
                {   keyplayer = 1;
                } else keyplayer = -1;
                if (keyplayer != -1)
                {   wormqueue(keyplayer, key[i].deltax, key[i].deltay);
            }   }
            elif (worm[1].lives > 0) // assumes key[which].special == TRAINER
            {   train(key[i].scancode);
}   }   }   }

EXPORT void resettime(void)
{   millielapsed = 0;
}

#define RUNDOWNX_0TH  27
#define RUNDOWNX_1ST  50
#define RUNDOWNX_2ND 150
#define RUNDOWNX_3RD 195 // "x"
#define RUNDOWNX_4TH 210
#define RUNDOWNX_5TH 255 // "="
#define RUNDOWNX_6TH 270

EXPORT void outro(void)
{   HBITMAP OldBitmapPtr;
    HFONT   OldFontPtr;
    TEXT    runstring[5][80 + 1],
            tempstring[12];
    int     cornerx, cornery,
            i, x, y,
            xoffset2;

    sprintf(runstring[0], "%s:", GetCatalogStr(CatalogPtr, MSG_LEVELBONUS , "Level Bonus" ));
    sprintf(runstring[1], "%s:", GetCatalogStr(CatalogPtr, MSG_NUMBERBONUS, "Number Bonus"));
    sprintf(runstring[2], "%s:", GetCatalogStr(CatalogPtr, MSG_TIMEBONUS  , "Time Bonus"  ));
    sprintf(runstring[3], "%s:", GetCatalogStr(CatalogPtr, MSG_DIMBONUS   , "Dim Bonus"   ));
    sprintf(runstring[4], "%s:", GetCatalogStr(CatalogPtr, MSG_GLOWBONUS  , "Glow Bonus"  ));

    // level just completed is level - 1

    xoffset2 = xoffset  + 48;
    cornerx  = xoffset2 - 20;
    cornery  = yoffset  - 18;

    rectfill_black(xoffset2 + 14, yoffset - 10, xoffset2 + 327, yoffset + 70);

    for (x = 3; x < 29; x++)
    {   raw_draw(cornerx + ( x * SQUAREX), cornery                , (UWORD) (GG_W_E + (showing * 58)));
        raw_draw(cornerx + ( x * SQUAREX), cornery + (7 * SQUAREY), (UWORD) (GG_W_E + (showing * 58)));
    }
    for (y = 1; y < 7; y++)
    {   raw_draw(cornerx + ( 2 * SQUAREX), cornery + (y * SQUAREY), (UWORD) (GG_N_S + (showing * 58)));
        raw_draw(cornerx + (29 * SQUAREX), cornery + (y * SQUAREY), (UWORD) (GG_N_S + (showing * 58)));
    }
    raw_draw(    cornerx + ( 2 * SQUAREX), cornery                , (UWORD) (GG_E_S + (showing * 58)));
    raw_draw(    cornerx + ( 2 * SQUAREX), cornery + (7 * SQUAREY), (UWORD) (GG_N_E + (showing * 58)));
    raw_draw(    cornerx + (29 * SQUAREX), cornery                , (UWORD) (GG_W_S + (showing * 58)));
    raw_draw(    cornerx + (29 * SQUAREX), cornery + (7 * SQUAREY), (UWORD) (GG_N_W + (showing * 58)));

    raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset                ,          TREASURE                               );
    if (worm[showing].numbers == 0)
    {   raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset + SQUAREY  , (UWORD) (FIRSTPAIN   + showing                  ));
    } else
    {   raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset + SQUAREY  , (UWORD) (FIRSTNUMBER + worm[showing].numbers - 1));
    }
    raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset + (SQUAREY * 2),          CLOCK                                  );
    raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset + (SQUAREY * 3), (UWORD) (FIRSTTAIL   + showing                  ));
    raw_draw(xoffset2 + RUNDOWNX_0TH, yoffset + (SQUAREY * 4), (UWORD) (FIRSTGLOW   + showing                  ));

    OnScreenRastPort = GetDC(MainWindowPtr);
    OldFontPtr = SelectObject(OnScreenRastPort, hFont);
    SetTextColor(OnScreenRastPort, worm[showing].colour);
    SetBkColor(OnScreenRastPort, BLACK);

    for (i = 0; i <= 4; i++)
    {   dotext(xoffset2 + RUNDOWNX_1ST, yoffset + 1 + (SQUAREY * i), runstring[i]);
        if (i == 2)
        {   if (quantity[2][0])
            {   tempstring[0] = 48 + (quantity[2][0] / 60);
                tempstring[1] = ':';
                if (quantity[2][0] % 60)
                {   tempstring[2] = 48 + ((quantity[2][0] % 60) / 10);
                    tempstring[3] = 48 + ((quantity[2][0] % 60) % 10);
                } else
                {   tempstring[2] =
                    tempstring[3] = '0';
                }
                tempstring[4] = 0; // important for IBM-PC version
            } else
            {   strcpy(tempstring, "0:00");
        }   }
        else
        {   stcl_d(tempstring, quantity[i][0]);
        }
        dotext(xoffset2 + RUNDOWNX_2ND, yoffset + 1 + (SQUAREY * i), tempstring);
        dotext(xoffset2 + RUNDOWNX_3RD, yoffset + 1 + (SQUAREY * i), "x");
        stcl_d(tempstring, quantity[i][1]);
        dotext(xoffset2 + RUNDOWNX_4TH, yoffset + 1 + (SQUAREY * i), tempstring);
        dotext(xoffset2 + RUNDOWNX_5TH, yoffset + 1 + (SQUAREY * i), "=");
        stcl_d(tempstring, quantity[i][2]);
        dotext(xoffset2 + RUNDOWNX_6TH, yoffset + 1 + (SQUAREY * i), tempstring);
    }

    SelectObject(OnScreenRastPort, OldFontPtr);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;

    stopfx();
}

EXPORT void say(STRPTR sentence, COLOUR colour)
{   HFONT OldFontPtr;
    SBYTE length = (SBYTE) strlen(sentence);

    if (!MainWindowPtr || alwaysrq)
    {   MessageBox(NULL, sentence, "Worm Wars: Message", MB_ICONEXCLAMATION | MB_OK);
        return;
    }

    // these lines are for the benefit of OnDraw()
    strcpy(currentsentence, sentence);
    currentcolour = colour;
    // truncate message
    if (length > SAYLIMIT)
    {   *(sentence + SAYLIMIT - 1) =
        *(sentence + SAYLIMIT - 2) =
        *(sentence + SAYLIMIT - 3) = '.';
        *(sentence + SAYLIMIT) = 0;
        length = SAYLIMIT;
    }

    OnScreenRastPort = GetDC(MainWindowPtr);
    // erase old message
    SelectObject(OnScreenRastPort, GetStockObject(BLACK_PEN));
    // both of these SelectObject()s are needed!
    SelectObject(OnScreenRastPort, GetStockObject(BLACK_BRUSH));
    if (isometric && a != LEVELEDIT)
    {   Rectangle
        (   OnScreenRastPort,
            STARTXPIXEL + (fieldy * SQUAREX / 2) + (SQUAREX / 2),
            0,
            ENDXPIXEL + (fieldy * SQUAREX / 2) + (SQUAREX / 2),
            STARTYPIXEL
        );
    } else
    {   Rectangle
        (   OnScreenRastPort,
            STARTXPIXEL,
            0,
            ENDXPIXEL,
            STARTYPIXEL
        );
    }
    // print new message
    OldFontPtr = SelectObject(OnScreenRastPort, hFont);
    SetTextColor(OnScreenRastPort, colour);
    SetBkColor(OnScreenRastPort, BLACK);
    if (isometric && a != LEVELEDIT)
    {   dotext
        (   STARTXPIXEL + (fieldy * SQUAREX / 2) + (SQUAREX / 2) + ((ENDXPIXEL - STARTXPIXEL) / 2) - (strlen(sentence) * FONTX / 4),
            3,
            sentence
        );
    } else
    {   dotext
        (   STARTXPIXEL + ((ENDXPIXEL - STARTXPIXEL) / 2) - (strlen(sentence) * FONTX / 4),
            3,
            sentence
        );
    }

    SelectObject(OnScreenRastPort, OldFontPtr);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void printstat(int player, int theline)
{   HFONT OldFontPtr;
    FLAG  release;

    if (OnScreenRastPort)
    {   release = FALSE;
    } else
    {   release = TRUE;
        OnScreenRastPort = GetDC(MainWindowPtr);
    }

    // erase old message
    SelectObject(OnScreenRastPort, GetStockObject(BLACK_BRUSH));
    // both of these SelectObject()s are needed!
    SelectObject(OnScreenRastPort, GetStockObject(BLACK_PEN));
    Rectangle
    (   OnScreenRastPort,
        14,
        20 + (theline       * FONTY) + (player * 120),
        74,
        20 + ((theline + 1) * FONTY) + (player * 120)
    );
    DISCARD SetBkColor(OnScreenRastPort, BLACK);
    SetTextColor(OnScreenRastPort, statcolour);
    OldFontPtr = SelectObject(OnScreenRastPort, hFont);
    dotext
    (   16,
        20 + (theline * FONTY) + (player * 120),
        stattext
    );
    SelectObject(OnScreenRastPort, OldFontPtr);

    if (release)
    {   DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
        OnScreenRastPort = NULL;
}   }

EXPORT void stcl_d(STRPTR out, long integer)
{   _ltoa(integer, out, 10);
}

MODULE void titlescreen(void)
{   TEXT  titlestring[11 + 40 + 1];
    RECT  therect;
    int   i,
          leftx, topy;

    for (;;)
    {   a = TITLESCREEN;
        clearkybd();

#ifdef TESTING
        worm[0].control =
        worm[1].control =
        worm[2].control =
        worm[3].control = THEAMIGA;
        a          = PLAYGAME;
        difficulty = NORMAL;
#else
        DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_TITLESCREEN), NULL, TitleDlgProc);
#endif

        if (a == LEVELEDIT)
        {   MenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(langs[language].menu));
            for (i = 0; i < MENUBITMAPS; i++)
            {   DISCARD SetMenuItemBitmaps
                (   MenuPtr,         // handle of menu
                    mib[i].menuitem, // menu item to receive new bitmaps
                    MF_BYCOMMAND,    // menu item flags
                    mib[i].bitmap,   // handle of unchecked bitmap
                    NULL             // handle of checked bitmap
                );
            }
            sprintf(titlestring, "Worm Wars: %s", GetCatalogStr(CatalogPtr, MSG_LEVELEDITOR, "Level Editor"));
        } else
        {   a = PLAYGAME;
            MenuPtr = NULL;
            strcpy(titlestring, "Worm Wars 9 for Windows");
        }
        calcsize();
        SystemParametersInfo(SPI_GETWORKAREA, 0, &therect, 0);
        leftx = (therect.right  - therect.left + 1 - DisplayWidth ) / 2;
        if (leftx < 0) leftx = 0;
        topy  = (therect.bottom - therect.top  + 1 - DisplayHeight) / 2;
        if (topy  < 0) topy  = 0;
        if (!(MainWindowPtr = CreateWindowEx
        (   0,
            g_szClassName,
            titlestring,
            WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
            leftx,
            topy,
            DisplayWidth, DisplayHeight,
            NULL,
            MenuPtr,
            InstancePtr,
            NULL
        )))
        {   MessageBox(NULL, "Window creation failed!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
            cleanexit(EXIT_FAILURE);
        }
        CurrentWindowPtr = MainWindowPtr;
        if (lame)
        {   SendMessage(MainWindowPtr, WM_CHANGEUISTATE, UISF_HIDEACCEL << 16 | UIS_CLEAR, 0); // this shows the keyboard shortcuts for the menu
        }
        do_systemmenu(MainWindowPtr);
        ShowWindow(MainWindowPtr, SW_SHOW);
        // SetForegroundWindow(MainWindowPtr);
        // InvalidateRect(MainWindowPtr, NULL, FALSE);
        // UpdateWindow(MainWindowPtr);

        if (a == PLAYGAME)
        {   ignorepaint = TRUE;
            newgame();
        } else
        {   assert(a == LEVELEDIT);
            fieldedit();
        }

        DestroyWindow(MainWindowPtr);
        MainWindowPtr = NULL;
}   }

EXPORT void waitasec(void)
{   Sleep(1000);
    clearkybd();
}

EXPORT void timing(void)
{   TRANSIENT ULONG time;
    PERSIST   ULONG oldtime = 0;

    do
    {   if (!turbo)
        {   do
            {   time = timeGetTime();
            } while (time < oldtime + (delay / 1000));
        }
        oldtime = time;

        if (a == PLAYGAME)
        {   millielapsed += delay / 1000;
            secondsleft = secondsperlevel - (millielapsed / 1000);
            // secondsleft could be negative in some instances. This is
            // handled in timeloop().
            r++;
            if (secondsleft != oldsecondsleft)
            {   timeloop();
                oldsecondsleft = secondsleft;
        }   }

        gameloop();
        ignorepaint = FALSE;
    } while (a == PLAYGAME);
}

EXPORT void clearkybd(void)
{   // MSG Message;

    ctrl = shift = FALSE;

    // while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE));
}

EXPORT FLAG busyloop(FLAG any)
{   /* When Worm Wars is forced to halt execution of the code, to wait
    on a user response, this is the routine that is called. It handles
    messages in a loop, and returns when done (FALSE for Esc, TRUE
    otherwise).

    If any is TRUE, any key is accepted. If FALSE, only P is accepted.

    The scancode is held in the low byte of the high word
    (ie. bits 16-23). Full description of Message.lParam:

[---High word--][---Low word---]
[H byte][L byte][H byte][L byte]
33222222222211111111110000000000
01987654321098765432109876543210

0*0[--]*[scanco][-repeat count-]
||| |  |
||| |  extended key flag
||| reserved
||context code (always 0 for WM_KEYDOWN)
|previous key state
transition state (always 0 for WM_KEYDOWN)

    Taking the bits 16-23 and zeroing the rest leaves us with a number which
    needs dividing by 65536 before it can be interpreted as a scancode. */

    FLAG        done = FALSE;
    MSG         Message;
    ULONG       scancode;
    PAINTSTRUCT ps;

    if (!MainWindowPtr)
    {   return TRUE;
    }

    globalabort = FALSE;
    while (!done && !globalabort)
    {   Sleep(0);
        while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
        {   switch (Message.message)
            {
            case WM_PAINT:
                BeginPaint(MainWindowPtr, &ps);
                redrawscreen();
                EndPaint(MainWindowPtr, &ps);
            acase WM_LBUTTONDOWN:
                if (any)
                {   done = TRUE;
                }
            acase WM_KEYDOWN:
                if (!(Message.lParam & 0x40000000)) // ie. if not a repeated key
                {   scancode = PARAM2SCANCODE(Message.lParam);
                    switch (scancode)
                    {
                    case SCAN_ESCAPE:
                        return FALSE;
                    acase SCAN_F:
                    case SCAN_M:
                        toggle((SBYTE) scancode);
                        if (any)
                        {   done = TRUE;
                        }
                    acase SCAN_F12:
                        if (ctrl)
                        {   iconify();
                        }
                    acase SCAN_LCTRL:
                    case SCAN_RCTRL:
                        ctrl = TRUE;
                    acase SCAN_LSHIFT:
                    case SCAN_RSHIFT:
                        shift = TRUE;
                    acase SCAN_P:
                    case SCAN_PAUSE:
                        done = TRUE;
                    adefault:
                        if (any)
                        {   done = TRUE;
                }   }   }
            acase WM_KEYUP:
                scancode = PARAM2SCANCODE(Message.lParam);
                switch (scancode)
                {
                case SCAN_LCTRL:
                case SCAN_RCTRL:
                    ctrl = FALSE;
                acase SCAN_LSHIFT:
                case SCAN_RSHIFT:
                    shift = FALSE;
                }
            adefault:
                TranslateMessage(&Message);
                DispatchMessage(&Message);
        }   }
        if (firebutton())
        {   done = TRUE;
    }   }

    return TRUE;
}

EXPORT void systemsetup(void)
{   currentsentence[0] = EOS;
}

EXPORT void toggle(SBYTE key)
{   if (key == SCAN_F)
    {   if (soundfx)
        {   stopfx(); // order-dependent
            soundfx = FALSE;
        } else
        {   soundfx = TRUE;
    }   }
    elif (key == SCAN_M)
    {   if (music)
        {   BASS_Stop();
            music = FALSE;
        } else
        {   music = TRUE;
            playsong(song);
    }   }

    soundicons();
}

MODULE void next(HWND hwnd)
{   DISCARD EnableWindow(GetDlgItem(hwnd, IDC_CLEARHISCORES), hiscoresareclear() ? FALSE : TRUE);

    if (hiscore[difficulty][0].fresh)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT1), EM_SETSEL, (WPARAM) 0, (LPARAM) -1);
        SetFocus(GetDlgItem(hwnd, IDC_EDIT1));
    } elif (hiscore[difficulty][1].fresh)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT2), EM_SETSEL, (WPARAM) 0, (LPARAM) -1);
        SetFocus(GetDlgItem(hwnd, IDC_EDIT2));
    } elif (hiscore[difficulty][2].fresh)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT3), EM_SETSEL, (WPARAM) 0, (LPARAM) -1);
        SetFocus(GetDlgItem(hwnd, IDC_EDIT3));
    } elif (hiscore[difficulty][3].fresh)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT4), EM_SETSEL, (WPARAM) 0, (LPARAM) -1);
        SetFocus(GetDlgItem(hwnd, IDC_EDIT4));
    } elif (hiscore[difficulty][4].fresh)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT5), EM_SETSEL, (WPARAM) 0, (LPARAM) -1);
        SetFocus(GetDlgItem(hwnd, IDC_EDIT5));
    } else
    {   SetFocus(GetDlgItem(hwnd, IDC_START));
        playsong(SONG_TITLE);
}   }

EXPORT void datestamp(void)
{   SYSTEMTIME  TheTime;
    TEXT        hourstring[3], minstring[3], mdaystring[3],
                monthstring[3], yearstring[5];

    GetLocalTime(&TheTime);

    stcl_d(hourstring, (int) TheTime.wHour);
    stcl_d(minstring, (int) TheTime.wMinute);
    align(minstring, 2, '0');
    minstring[2] = 0;
    stcl_d(mdaystring, (int) TheTime.wDay);
    stcl_d(monthstring, (int) TheTime.wMonth);
    stcl_d(yearstring, (int) TheTime.wYear);
    yearstring[0] = yearstring[2];
    yearstring[1] = yearstring[3];
    yearstring[2] = 0;

    strcpy(times, hourstring);
    strcat(times, ":");
    strcat(times, minstring);
    strcpy(date, mdaystring);
    strcat(date, "/");
    strcat(date, monthstring);
    strcat(date, "/");
    strcat(date, yearstring);
}

EXPORT void turborender(void)
{   int player,
        x, y;

    assert(a != LEVELEDIT);

    if (!level || !shuffle)
    {   sourcelevel = level;
    }

    changefield();

    for (x = 0; x <= fieldx; x++)
    {   for (y = 0; y <= fieldy; y++)
        {   draw(x, y, field[x][y]);
    }   }
    for (y = 0; y <= fieldy; y++)
    {   draw(fieldx + 1, y, BLACKARROW);
}   }

MODULE void soundicons(void)
{   // visually updates the 'sound' indicator

    if (soundfx && music)
    {   draw(MUSICICONX, ICONY, BOTH);
    } elif (music)
    {   draw(MUSICICONX, ICONY, MUSIC);
    } elif (soundfx)
    {   draw(MUSICICONX, ICONY, FX);
    } else
    {   draw(MUSICICONX, ICONY, BLACKENED);
    }

    if (turbo)
    {   draw(CLOCKICONX, ICONY, CLOCK);
    } else
    {   draw(CLOCKICONX, ICONY, BLACKENED);
}   }

EXPORT void whiteline(void)
{   statcolour = WHITE;
}
EXPORT void colourline(int player)
{   statcolour = worm[player].colour;
}

EXPORT void playsong(UBYTE whichsong)
{   assert(whichsong >= 0 && whichsong < SONGS);
    if (music && musicable)
    {   BASS_Stop();
        BASS_Start();
        if (song == whichsong)
        {   BASS_ChannelPlay(SongHandle[whichsong], FALSE); // play from where last stopped
        } else
        {   BASS_ChannelPlay(SongHandle[whichsong], TRUE); // play from beginning
    }   }
    song = whichsong;
}

MODULE void load_music(void)
{   if (BASS_Init
    (   -1,
        44100, // does it really need to be this high?
        0,
        AnchorWindowPtr,
        NULL
    ))
    {   if
        (   (SongHandle[0] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\TITLE.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[1] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\EDITOR.MOD"  , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[2] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\BONUS.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[3] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAMEOVER.MOD", 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[4] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAME1.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[5] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAME2.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[6] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAME3.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[7] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAME4.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
         && (SongHandle[8] = (HMUSIC) BASS_MusicLoad(FALSE, "MUSIC\\GAME5.MOD"   , 0, 0, BASS_MUSIC_RAMP | BASS_MUSIC_LOOP, 0))
        )
        {   musicable = TRUE;
        } /* else
        {   MessageBox(NULL, "Can't load music!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
        } */
    } else
    {   MessageBox(NULL, "Can't initialize music subsystem!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
}   }

EXPORT void domusic(void)
{   soundicons();

    if (music)
    {   playsong(song);
    } else
    {   BASS_Stop();
}   }

MODULE void calcsize(void)
{   RECT therect;
    int  barheight,
         which;

    if (a == LEVELEDIT)
    {   fieldx        = MINFIELDX;
        fieldy        = MINFIELDY;
        DisplayWidth  = 626;
        DisplayHeight = 533 + GetSystemMetrics(SM_CYMENU);
        for (which = 0; which < PSEUDOGADGETS; which++)
        {   pseudotop[which] = FIELDCENTREY - 4 + (((which * 2) - (PSEUDOGADGETS - 1)) * SQUAREY);
    }   }
    else
    {   DisplayWidth  = resolutions[resolution].width;
        DisplayHeight = resolutions[resolution].height;
        SystemParametersInfo(SPI_GETWORKAREA, 0, &therect, 0);
        barheight = GetSystemMetrics(SM_CYSCREEN) - therect.bottom - therect.top; // 40 for Windows 7
        DisplayHeight -= barheight;
        fieldy      =
        coordfieldy = ((DisplayHeight - 61) / SQUAREY)    ;
        coordfieldx = ((DisplayWidth  - 88) / SQUAREX) - 2;
        if (isometric)
        {   fieldx  = ((DisplayWidth  - 88 - (SQUAREX / 2)) / SQUAREX) - 2 - (fieldy / 2);
        } else
        {   fieldx  = coordfieldx;
    }   }

    xoffset = STARTXPIXEL + (coordfieldx * SQUAREX / 2) - 200;
    yoffset = STARTYPIXEL + (coordfieldy * SQUAREY / 2) -  30;
}

MODULE void load_samples(void)
{   FILE* LocalHandle;
    int   i;

    for (i = 0; i < SAMPLES; i++)
    {   if
        (   ((samp[i].length = getsize(samp[i].filename)))
         && ((LocalHandle = fopen(samp[i].filename, "rb")))
        )
        {   SoundBuffer[i] = (UBYTE*) malloc(samp[i].length); // should really check return code
            DISCARD fread(SoundBuffer[i], samp[i].length, 1, LocalHandle); // should really check return code
            DISCARD fclose(LocalHandle);
            // LocalHandle = NULL;
            samp[i].hertz =  SoundBuffer[i][24]
                          + (SoundBuffer[i][25] *      256)
                          + (SoundBuffer[i][26] *    65536)
                          + (SoundBuffer[i][27] * 16777216);
}   }   }

EXPORT void effect(int index)
{   TRANSIENT WAVEFORMATEX l_WaveFormatEx;
    PERSIST   int          channel = 0;

    if (soundfx && SoundBuffer[index])
    {   if (channel == CHANNELS - 1)
        {   channel = 0;
        } else
        {   channel++;
        }

        if (WavePrepared[channel])
        {   DISCARD waveOutReset(hWave[channel]);
            DISCARD waveOutUnprepareHeader(hWave[channel], &WaveHdr[channel], sizeof(WAVEHDR));
            WavePrepared[channel] = FALSE;
        }
        if (WaveOpened[channel])
        {   DISCARD waveOutClose(hWave[channel]);
            WaveOpened[channel] = FALSE;
        }

        l_WaveFormatEx.wFormatTag      = WAVE_FORMAT_PCM;
        l_WaveFormatEx.nChannels       = 1;
        l_WaveFormatEx.nSamplesPerSec  =
        l_WaveFormatEx.nAvgBytesPerSec = samp[index].hertz;
        l_WaveFormatEx.wBitsPerSample  = 8;
        l_WaveFormatEx.nBlockAlign     = 1;
        l_WaveFormatEx.nAvgBytesPerSec = l_WaveFormatEx.nSamplesPerSec * l_WaveFormatEx.nBlockAlign;
        l_WaveFormatEx.cbSize          = sizeof(WAVEFORMATEX); // needed for Windows 7!
        memset(&hWave[channel], 0, sizeof(HWAVEOUT));
        DISCARD waveOutOpen
        (   &hWave[channel],
            WAVE_MAPPER,
            &l_WaveFormatEx,
            0,
            0,
            CALLBACK_NULL
        );
        WaveOpened[channel] = TRUE;

        WaveHdr[channel].lpData          = (char*) (SoundBuffer[index] + 44);
        WaveHdr[channel].dwBufferLength  = samp[index].length - 44;
        WaveHdr[channel].dwBytesRecorded = 0;
        WaveHdr[channel].dwUser          = 0;
        WaveHdr[channel].dwFlags         = 0;
        WaveHdr[channel].dwLoops         = 1;
        WaveHdr[channel].lpNext          = NULL;
        WaveHdr[channel].reserved        = (unsigned long) NULL;
        DISCARD waveOutPrepareHeader(hWave[channel], &WaveHdr[channel], sizeof(WAVEHDR));
        WavePrepared[channel] = TRUE;

        WaveHdr[channel].dwFlags         |= WHDR_BEGINLOOP | WHDR_ENDLOOP;
        DISCARD waveOutWrite(hWave[channel], &WaveHdr[channel], sizeof(WAVEHDR));
}   }

EXPORT void stopfx(void)
{   int i;

    for (i = 0; i < CHANNELS; i++)
    {   if (WavePrepared[i])
        {   DISCARD waveOutReset(hWave[i]);
            DISCARD waveOutUnprepareHeader(hWave[i], &WaveHdr[i], sizeof(WAVEHDR));
            WavePrepared[i] = FALSE;
}   }   }

MODULE ULONG getsize(STRPTR filename)
{   HANDLE hFile /* = NULL */ ;
    ULONG  size;

    hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
    if (hFile == INVALID_HANDLE_VALUE)
    {   return 0;
    }
    size = GetFileSize(hFile, NULL);
    CloseHandle(hFile);
    // hFile = NULL;

    if (size == (ULONG) -1)
    {   return 0;
    }

    return size;
}

MODULE void redrawscreen(void)
{   SBYTE x, y;
    UBYTE i;
    int   player;

    clearscreen();
    say(currentsentence, currentcolour);

    if (a == LEVELEDIT)
    {   le_dosidebar();
    }

    if (showing >= 0 && showing <= 3)
    {   outro();
    } elif (showing == 5)
    {   intro();
        repaint_intro();
    } elif (a == LEVELEDIT)
    {   le_drawfield();
    } elif ((a == PLAYGAME || a == GAMEOVER) && level <= MAXLEVELS)
    {   for (x = 0; x <= fieldx; x++)
        {   for (y = 0; y <= fieldy; y++)
            {   draw(x, y, gfxfield[x][y]);
        }   }
        if (a != LEVELEDIT)
        {   for (y = 0; y <= fieldy; y++)
            {   updatearrow(y);
}   }   }   }

MODULE BOOL CALLBACK AboutDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   HKEY hKey;
    int  i, j,
         thesize = MAX_PATH;

    switch (Message)
    {
    case WM_INITDIALOG:
        DISCARD SetWindowText( hwnd, GetCatalogStr(CatalogPtr, MSG_HAIL_ABOUT, "About Worm Wars"));
        DISCARD SetDlgItemText(hwnd, IDL_ABOUT         , TITLEBAR " for Windows");
        DISCARD SetDlgItemText(hwnd, IDL_RELEASEDATE   , RELEASEDATE_LONG);
        setdlgtext(            hwnd, IDC_SYSINFO       , MSG_SYSTEM_INFORMATION, "&System information...");
        DISCARD SetDlgItemText(hwnd, IDL_TRANSLATOR    , GetCatalogStr(CatalogPtr, MSG_TRANSLATOR, "English translation by James Jacobs"));
        DISCARD SendMessage(GetDlgItem(hwnd, IDL_ABOUT), WM_SETFONT, (WPARAM) hBoldFont, MAKELPARAM(TRUE, 0));
        SetFocus(GetDlgItem(hwnd, IDOK));

        if (RegOpenKeyEx
        (   HKEY_LOCAL_MACHINE,
            "SOFTWARE\\Microsoft\\Shared Tools\\MSInfo",
            0,
            KEY_READ,
            &hKey
        ) == ERROR_SUCCESS)
        {   if (RegQueryValueEx
            (   hKey,
                "Path",
                0,
                NULL,
                SysInfoName,
                (LPDWORD) &thesize
            ) == ERROR_SUCCESS && SysInfoName[0])
            {   EnableWindow(GetDlgItem(hwnd, IDC_SYSINFO), TRUE);
            }
            DISCARD RegCloseKey(hKey);
        }

        effect(FXHELP);
        DISCARD SetFocus(GetDlgItem(hwnd, IDOK));
        return FALSE; // must be FALSE so that our SetFocus()sing is respected
    case WM_DRAWITEM:
        drawitem((struct tagDRAWITEMSTRUCT*) lParam);
        return TRUE;
    case WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        SetForegroundWindow(TitleWindowPtr);
        return TRUE;
    acase WM_KEYDOWN:
        if (PARAM2SCANCODE(lParam) == SCAN_ESCAPE) // this doesn't work for some reason
        {   clearkybd();
            EndDialog(hwnd, 0);
            SetForegroundWindow(TitleWindowPtr);
            return TRUE;
        }
    acase WM_COMMAND:
        switch (LOWORD(wParam))
        {
        case IDC_LINK1:
        case IDC_AMIGANSOFTWARE:
            openurl("http://amigan.1emu.net/releases/");
        acase IDC_LINK2:
            ShellExecute(0, NULL, "mailto:amigansoftware@gmail.com", NULL, NULL, SW_SHOW);
        acase IDC_SYSINFO:
            DISCARD ShellExecute(0, NULL, SysInfoName, NULL, NULL, SW_SHOW);
        acase IDOK:
            clearkybd();
            EndDialog(hwnd, 0);
            SetForegroundWindow(TitleWindowPtr);
        }
        return TRUE;
    adefault:
        return FALSE;
    }
    return TRUE;
}

MODULE BOOL CALLBACK CreaturesDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   TRANSIENT int  i, j;
    FAST      TEXT tempstring[3 + 1];

    switch (Message)
    {
    case WM_INITDIALOG:
        DISCARD SetWindowText( hwnd,              GetCatalogStr(CatalogPtr, MSG_HAIL_CREATURES, "Creatures"));

        j = 0;
        for (i = 0; i <= SPECIES; i++)
        {   if (creatureinfo[i].user)
            {   DISCARD SetDlgItemText(hwnd, IDL_CREATURE1 + j, creatureinfo[i].name);
                sprintf(tempstring, "%d", creatureinfo[i].score);
                DISCARD SetDlgItemText(hwnd, IDL_PTS1      + j, tempstring);
                j++;
        }   }

        DISCARD SetDlgItemText(hwnd, IDL_POINTS1, GetCatalogStr(CatalogPtr, MSG_POINTS3,        "Points"));
        DISCARD SetDlgItemText(hwnd, IDL_POINTS2, GetCatalogStr(CatalogPtr, MSG_POINTS3,        "Points"));

        return TRUE;
    case WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        SetForegroundWindow(TitleWindowPtr);
        return TRUE;
    case WM_KEYDOWN:
        if (PARAM2SCANCODE(lParam) == SCAN_ESCAPE) // this doesn't work for some reason
        {   clearkybd();
            EndDialog(hwnd, 0);
            SetForegroundWindow(TitleWindowPtr);
            return TRUE;
        }
        return FALSE;
    default:
        return FALSE;
}   }

MODULE BOOL CALLBACK FruitsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int i;

    switch (Message)
    {
    case WM_INITDIALOG:
        DISCARD SetWindowText( hwnd,                  GetCatalogStr(CatalogPtr, MSG_HAIL_FRUITS,      "Fruits"));

        DISCARD SetDlgItemText(hwnd, IDL_POINTS1,     GetCatalogStr(CatalogPtr, MSG_POINTS3,          "Points"));

        for (i = FIRSTFRUIT; i <= LASTFRUIT; i++)
        {   DISCARD SetDlgItemText(hwnd, i - FIRSTFRUIT + IDL_FRUIT1, fruitname[i - FIRSTFRUIT]);
        }

        return TRUE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        SetForegroundWindow(TitleWindowPtr);
        return TRUE;
    acase WM_KEYDOWN:
        if (PARAM2SCANCODE(lParam) == SCAN_ESCAPE) // this doesn't work for some reason
        {   clearkybd();
            EndDialog(hwnd, 0);
            return TRUE;
        }
        return FALSE;
    adefault:
        return FALSE;
}   }

MODULE BOOL CALLBACK ObjectsDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   int i;

    switch (Message)
    {
    case WM_INITDIALOG:
        DISCARD SetWindowText(hwnd, GetCatalogStr(CatalogPtr, MSG_HAIL_OBJECTS, "Objects"));

        for (i = 0; i <= LASTOBJECT; i++)
        {    DISCARD SetDlgItemText(hwnd, IDL_OBJECT1 + i, objectinfo[i].desc);
        }

        return TRUE;
    acase WM_CLOSE:
        clearkybd();
        EndDialog(hwnd, 0);
        SetForegroundWindow(TitleWindowPtr);
        return TRUE;
    acase WM_KEYDOWN:
        if (PARAM2SCANCODE(lParam) == SCAN_ESCAPE) // this doesn't work for some reason
        {   clearkybd();
            EndDialog(hwnd, 0);
            SetForegroundWindow(TitleWindowPtr);
            return TRUE;
        }
        return FALSE;
    adefault:
        return FALSE;
}   }

MODULE BOOL CALLBACK TitleDlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{   TRANSIENT RECT  therect;
    TRANSIENT LONG  gid;
    TRANSIENT int   i;
    PERSIST   FLAG  first = TRUE,
                    opening,
                    startable,
                    ignore;
    PERSIST   LONG  focussed;

    switch (Message)
    {
    case WM_INITDIALOG:
        opening = TRUE;

        CurrentWindowPtr = TitleWindowPtr = hwnd;

        GetWindowRect(hwnd, &therect);
        DISCARD SetWindowPos
        (   hwnd,
            HWND_TOP,
            (GetSystemMetrics(SM_CXSCREEN) / 2) - ((therect.right  - therect.left + 1) / 2),
            (GetSystemMetrics(SM_CYSCREEN) / 2) - ((therect.bottom - therect.top  + 1) / 2),
            0, 0,
            SWP_NOSIZE | SWP_NOCOPYBITS
        );

        SendMessage(hwnd, WM_SETICON, ICON_BIG,   (LPARAM) BigIconPtr);
        SendMessage(hwnd, WM_SETICON, ICON_SMALL, (LPARAM) SmallIconPtr);
        do_systemmenu(hwnd);
        if (joys < 2)
        {   if (worm[3].control == JOYSTICK) worm[3].control = NONE;
            EnableWindow(GetDlgItem(hwnd, IDC_YELLOW_JOY), FALSE);
            if (joys < 1)
            {   if (worm[2].control == JOYSTICK) worm[2].control = NONE;
                EnableWindow(GetDlgItem(hwnd, IDC_BLUE_JOY), FALSE);
        }   }

        for (i = 1; i < RESOLUTIONS; i++)
        {   if (entirex < resolutions[i].width || entirey < resolutions[i].height)
            {   EnableWindow(GetDlgItem(hwnd, IDC_RES800 + i), FALSE);
        }   }

        if (resolution == 0)
        {   EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), FALSE);
        }

        renderhiscores(hwnd);

        switch (worm[0].control)
        {
        case NONE:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_GREEN_NONE,
                IDC_GREEN_KYBD,
                IDC_GREEN_NONE
            );
        acase IBMPC:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_GREEN_NONE,
                IDC_GREEN_KYBD,
                IDC_GREEN_IBM
            );
        acase KEYBOARD:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_GREEN_NONE,
                IDC_GREEN_KYBD,
                IDC_GREEN_KYBD
            );
        }
        switch (worm[1].control)
        {
        case NONE:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_RED_NONE,
                IDC_RED_KYBD,
                IDC_RED_NONE
            );
        acase IBMPC:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_RED_NONE,
                IDC_RED_KYBD,
                IDC_RED_IBM
            );
        acase KEYBOARD:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_RED_NONE,
                IDC_RED_KYBD,
                IDC_RED_KYBD
            );
        }
        switch (worm[2].control)
        {
        case NONE:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_BLUE_NONE,
                IDC_BLUE_JOY,
                IDC_BLUE_NONE
            );
        acase IBMPC:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_BLUE_NONE,
                IDC_BLUE_JOY,
                IDC_BLUE_IBM
            );
        acase JOYSTICK:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_BLUE_NONE,
                IDC_BLUE_JOY,
                IDC_BLUE_JOY
            );
        }
        switch (worm[3].control)
        {
        case NONE:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_YELLOW_NONE,
                IDC_YELLOW_JOY,
                IDC_YELLOW_NONE
            );
        acase IBMPC:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_YELLOW_NONE,
                IDC_YELLOW_JOY,
                IDC_YELLOW_IBM
            );
        acase JOYSTICK:
            DISCARD CheckRadioButton
            (   hwnd,
                IDC_YELLOW_NONE,
                IDC_YELLOW_JOY,
                IDC_YELLOW_JOY
            );
        }
        DISCARD CheckRadioButton
        (   hwnd,
            IDC_RES800,
            IDC_RES1280TALL,
            IDC_RES800 + resolution
        );
        DISCARD CheckRadioButton
        (   hwnd,
            IDC_ENG,
            IDC_SPA,
            IDC_ENG + language
        );
        switch (difficulty)
        {
        case  0: DISCARD CheckRadioButton(hwnd, IDC_EASY, IDC_VERYHARD, IDC_EASY    );
        acase 1: DISCARD CheckRadioButton(hwnd, IDC_EASY, IDC_VERYHARD, IDC_NORMAL  );
        acase 2: DISCARD CheckRadioButton(hwnd, IDC_EASY, IDC_VERYHARD, IDC_HARD    );
        acase 3: DISCARD CheckRadioButton(hwnd, IDC_EASY, IDC_VERYHARD, IDC_VERYHARD);
        }
        if (music)     SendMessage(GetDlgItem(hwnd, IDC_MUSIC    ), BM_SETCHECK, BST_CHECKED, 0);
        if (soundfx)   SendMessage(GetDlgItem(hwnd, IDC_EFFECTS  ), BM_SETCHECK, BST_CHECKED, 0);
        if (tdworms)   SendMessage(GetDlgItem(hwnd, IDC_3DWORMS  ), BM_SETCHECK, BST_CHECKED, 0);
        if (autosave)  SendMessage(GetDlgItem(hwnd, IDC_AUTOSAVE ), BM_SETCHECK, BST_CHECKED, 0);
        if (shuffle)   SendMessage(GetDlgItem(hwnd, IDC_SHUFFLE  ), BM_SETCHECK, BST_CHECKED, 0);
        if (engraved)  SendMessage(GetDlgItem(hwnd, IDC_ENGRAVED ), BM_SETCHECK, BST_CHECKED, 0);
        if (isometric) SendMessage(GetDlgItem(hwnd, IDC_ISOMETRIC), BM_SETCHECK, BST_CHECKED, 0);
        for (i = 0; i <= 4; i++)
        {   SetDlgItemText(hwnd, IDC_EDIT1 + i, hiscore[difficulty][i].name);
        }

        retranslate();
        next(hwnd);

        DISCARD SendMessage(GetDlgItem(hwnd, IDC_START), WM_SETFONT, (WPARAM) hBoldFont, MAKELPARAM(TRUE, 0));

        if (lame)
        {   SendMessage(hwnd, WM_CHANGEUISTATE, UISF_HIDEACCEL << 16 | UIS_CLEAR, 0); // this shows the keyboard shortcuts for the menu
        }

        DISCARD RegisterHotKey(hwnd, IDK_F12, MOD_CONTROL, VK_F12); // this will fail if there are multiple instances running
        // presumably this gets automatically freed by Windows when we destroy the window

        ignore = TRUE;
        opening = FALSE;

        focussed = -1;
        DISCARD SetFocus(GetDlgItem(hwnd, IDC_START));
        return FALSE; // must be FALSE so that our SetFocus()sing is respected
    acase WM_HOTKEY:
        iconify();
    acase WM_DRAWITEM:
        drawitem((struct tagDRAWITEMSTRUCT*) lParam);
        return TRUE;
    acase WM_CLOSE:
        if (verify())
        {   EndDialog(hwnd, 0);
            cleanexit(EXIT_SUCCESS);
        }
        return TRUE;
    acase WM_SYSCOMMAND:
        switch (LOWORD(wParam))
        {
        case ID_FILE_ICONIFY:
            iconify();
        acase ID_HELP_MANUAL:
            ShellExecute(0, NULL, "WORMWARS.HTM", NULL, NULL, SW_SHOW);
        acase ID_HELP_UPDATE:
            help_update();
        acase ID_HELP_ABOUT:
            DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, AboutDlgProc);
        }
        return 0;
    acase USR_NOTIFYICON:
        assert(iconified);
        switch (lParam)
        {
        case WM_LBUTTONDBLCLK:
            uniconify(FALSE);
        acase WM_RBUTTONUP:
            traymenu();
        }
        return 0;
    acase WM_COMMAND:
        if (opening)
        {   break;
        }
        switch (LOWORD(wParam))
        {
        case IDCANCEL: // eg. Esc key
            if (verify())
            {   EndDialog(hwnd, 0);
                cleanexit(EXIT_SUCCESS);
            }
        acase IDC_GREEN_NONE:
            worm[0].control = NONE;
        acase IDC_GREEN_IBM:
            worm[0].control = THEAMIGA;
        acase IDC_GREEN_KYBD:
            worm[0].control = KEYBOARD;
        acase IDC_RED_NONE:
            worm[1].control = NONE;
        acase IDC_RED_IBM:
            worm[1].control = THEAMIGA;
        acase IDC_RED_KYBD:
            worm[1].control = KEYBOARD;
        acase IDC_BLUE_NONE:
            worm[2].control = NONE;
        acase IDC_BLUE_IBM:
            worm[2].control = THEAMIGA;
        acase IDC_BLUE_JOY:
            worm[2].control = JOYSTICK;
        acase IDC_YELLOW_NONE:
            worm[3].control = NONE;
        acase IDC_YELLOW_IBM:
            worm[3].control = THEAMIGA;
        acase IDC_YELLOW_JOY:
            worm[3].control = JOYSTICK;
        acase IDC_3DWORMS:   tdworms   = (SendMessage(GetDlgItem(hwnd, IDC_3DWORMS  ), BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
        acase IDC_AUTOSAVE:  autosave  = (SendMessage(GetDlgItem(hwnd, IDC_AUTOSAVE ), BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
        acase IDC_SHUFFLE:   shuffle   = (SendMessage(GetDlgItem(hwnd, IDC_SHUFFLE  ), BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
        acase IDC_ENGRAVED:  engraved  = (SendMessage(GetDlgItem(hwnd, IDC_ENGRAVED ), BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
        acase IDC_ISOMETRIC: isometric = (SendMessage(GetDlgItem(hwnd, IDC_ISOMETRIC), BM_GETCHECK, 0, 0) == BST_CHECKED) ? TRUE : FALSE;
        acase IDC_EASY:
            difficulty = DIFFICULTY_EASY;
            renderhiscores(hwnd);
        acase IDC_NORMAL:
            difficulty = DIFFICULTY_NORMAL;
            renderhiscores(hwnd);
        acase IDC_HARD:
            difficulty = DIFFICULTY_HARD;
            renderhiscores(hwnd);
        acase IDC_VERYHARD:
            difficulty = DIFFICULTY_VERYHARD;
            renderhiscores(hwnd);
        acase IDC_RES800:
            resolution = 0;
            if (isometric)
            {   isometric = FALSE;
                SendMessage(GetDlgItem(hwnd, IDC_ISOMETRIC), BM_SETCHECK, BST_UNCHECKED, 0);
                EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), FALSE);
            }
        acase IDC_RES1024SHORT:
            if (entirex >= resolutions[1].width && entirey >= resolutions[1].height)
            {   resolution = 1;
                EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), TRUE);
            } else
            {   DISCARD CheckRadioButton
                (   hwnd,
                    IDC_RES800,
                    IDC_RES1280TALL,
                    IDC_RES800 + resolution
                );
                UpdateWindow(GetDlgItem(hwnd, IDC_RES800 + resolution));
            }
        acase IDC_RES1024TALL:
            if (entirex >= resolutions[2].width && entirey >= resolutions[2].height)
            {   resolution = 2;
                EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), TRUE);
            } else
            {   DISCARD CheckRadioButton
                (   hwnd,
                    IDC_RES800,
                    IDC_RES1280TALL,
                    IDC_RES800 + resolution
                );
                UpdateWindow(GetDlgItem(hwnd, IDC_RES800 + resolution));
            }
        acase IDC_RES1280SHORT:
            if (entirex >= resolutions[3].width && entirey >= resolutions[3].height)
            {   resolution = 3;
                EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), TRUE);
            } else
            {   DISCARD CheckRadioButton
                (   hwnd,
                    IDC_RES800,
                    IDC_RES1280TALL,
                    IDC_RES800 + resolution
                );
                UpdateWindow(GetDlgItem(hwnd, IDC_RES800 + resolution));
            }
        acase IDC_RES1280TALL:
            if (entirex >= resolutions[4].width && entirey >= resolutions[4].height)
            {   resolution = 4;
                EnableWindow(GetDlgItem(hwnd, IDC_ISOMETRIC), TRUE);
            } else
            {   DISCARD CheckRadioButton
                (   hwnd,
                    IDC_RES800,
                    IDC_RES1280TALL,
                    IDC_RES800 + resolution
                );
                UpdateWindow(GetDlgItem(hwnd, IDC_RES800 + resolution));
            }
        acase IDC_ENG:
            freelanguage();
            language = LANGUAGE_ENG;
            changelanguage();
        acase IDC_GER:
            freelanguage();
            language = LANGUAGE_GER;
            changelanguage();
        acase IDC_GRE:
            freelanguage();
            language = LANGUAGE_GRE;
            changelanguage();
        acase IDC_ITA:
            freelanguage();
            language = LANGUAGE_ITA;
            changelanguage();
        acase IDC_POL:
            freelanguage();
            language = LANGUAGE_POL;
            changelanguage();
        acase IDC_SPA:
            freelanguage();
            language = LANGUAGE_SPA;
            changelanguage();
        acase IDC_EFFECTS:
            if (SendMessage(GetDlgItem(hwnd, IDC_EFFECTS), BM_GETCHECK, 0, 0) == BST_CHECKED)
            {   soundfx = TRUE;
            } else
            {   soundfx = FALSE;
            }
        acase IDC_MUSIC:
            if (!music && SendMessage(GetDlgItem(hwnd, IDC_MUSIC), BM_GETCHECK, 0, 0) == BST_CHECKED)
            {   music = TRUE;
                domusic();
            } elif (music && SendMessage(GetDlgItem(hwnd, IDC_MUSIC), BM_GETCHECK, 0, 0) != BST_CHECKED)
            {   music = FALSE;
                domusic();
            }
        acase IDC_CREATURES:
            DISCARD EnableWindow(hwnd, FALSE);
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CREATURES), TitleWindowPtr, CreaturesDlgProc);
            DISCARD EnableWindow(hwnd, TRUE);
        acase IDC_OBJECTS:
            DISCARD EnableWindow(hwnd, FALSE);
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_OBJECTS  ), TitleWindowPtr, ObjectsDlgProc);
            DISCARD EnableWindow(hwnd, TRUE);
        acase IDC_FRUITS:
            DISCARD EnableWindow(hwnd, FALSE);
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_FRUITS   ), TitleWindowPtr, FruitsDlgProc);
            DISCARD EnableWindow(hwnd, TRUE);
        acase IDC_LOAD:
            fileopen(FALSE);
        acase IDC_SAVE:
            /* if (pathname[0])
            {   filesaveas(FALSE);
            } else */
            filesaveas(TRUE);
        acase IDC_START:
            a = PLAYGAME;
            clearkybd();
            EndDialog(hwnd, 0);
        acase IDC_LEVELEDITOR:
            a = LEVELEDIT;
            clearkybd();
            EndDialog(hwnd, 0);
        acase IDOK: // eg. Enter key
            // GetFocus() is always returning NULL under Windows XP, unfortunately.
            if
            (   (focussed == -1 || GetFocus() == GetDlgItem(hwnd, IDC_START))
             && startable
            )
            {   clearkybd();
                EndDialog(hwnd, 0);
            } elif (focussed >= 0 && focussed <= 4)
            {   killfocus(hwnd, focussed);
            }
        acase IDC_EDIT1:
        case IDC_EDIT2:
        case IDC_EDIT3:
        case IDC_EDIT4:
        case IDC_EDIT5:
            if (HIWORD(wParam) == EN_KILLFOCUS)
            {   if (ignore)
                {   ignore = FALSE;
                    next(hwnd);
                } else
                {   if (LOWORD(wParam) >= IDC_EDIT1 && LOWORD(wParam) <= IDC_EDIT5)
                    {   focussed = LOWORD(wParam) - IDC_EDIT1;
                        killfocus(hwnd, focussed);
            }   }   }
            elif (HIWORD(wParam) == EN_SETFOCUS)
            {   if (LOWORD(wParam) >= IDC_EDIT1 && LOWORD(wParam) <= IDC_EDIT5)
                {   focussed = LOWORD(wParam) - IDC_EDIT1;
                    if (!hiscore[difficulty][focussed].fresh)
                    {   killfocus(hwnd, focussed);
            }   }   }
        acase IDC_MANUAL:
            ShellExecute(0, NULL, "WORMWARS.HTM", NULL, NULL, SW_SHOW);
        acase IDC_UPDATE:
            help_update();
        acase IDC_ABOUT:
            DISCARD EnableWindow(hwnd, FALSE);
            DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_ABOUT), TitleWindowPtr, AboutDlgProc);
            DISCARD EnableWindow(hwnd, TRUE);
        acase IDC_CLEARHISCORES:
            scores_modified = TRUE;
            clearhiscores();
            renderhiscores(hwnd);
            // assert(hiscoresareclear());
            EnableWindow(GetDlgItem(hwnd, IDC_CLEARHISCORES), FALSE);
        acase ID_FILE_UNICONIFY: // needed!
            uniconify(FALSE);
        acase ID_FILE_EXITDOS:
            cleanexit(EXIT_SUCCESS);
     /* acase ID_OPTIONS_PAUSED:
            assert(0); */
        }

        if
        (   SendMessage(GetDlgItem(hwnd, IDC_GREEN_NONE ), BM_GETCHECK, 0, 0) == BST_CHECKED
         && SendMessage(GetDlgItem(hwnd, IDC_RED_NONE   ), BM_GETCHECK, 0, 0) == BST_CHECKED
         && SendMessage(GetDlgItem(hwnd, IDC_BLUE_NONE  ), BM_GETCHECK, 0, 0) == BST_CHECKED
         && SendMessage(GetDlgItem(hwnd, IDC_YELLOW_NONE), BM_GETCHECK, 0, 0) == BST_CHECKED
        )
        {   startable = FALSE;
            EnableWindow(GetDlgItem(hwnd, IDC_START), FALSE);
        } else
        {   startable = TRUE;
            EnableWindow(GetDlgItem(hwnd, IDC_START), TRUE);
        }
        return TRUE;
    case WM_CTLCOLORSTATIC: // no need for acase
        gid = GetWindowLong((HWND) lParam, GWL_ID);
        switch (gid)
        {
        case IDC_LEVEL1:
        case IDC_LEVEL2:
        case IDC_LEVEL3:
        case IDC_LEVEL4:
        case IDC_LEVEL5:
            if (hiscore[difficulty][gid - IDC_LEVEL1].player != -1)
            {   SetBkColor((HDC) wParam, BGRColour[hiscore[difficulty][gid - IDC_LEVEL1].player]);
                return (LRESULT) hBrush[hiscore[difficulty][gid - IDC_LEVEL1].player];
            }
        acase IDC_SCORE1:
        case IDC_SCORE2:
        case IDC_SCORE3:
        case IDC_SCORE4:
        case IDC_SCORE5:
            if (hiscore[difficulty][gid - IDC_SCORE1].player != -1)
            {   SetBkColor((HDC) wParam, BGRColour[hiscore[difficulty][gid - IDC_SCORE1].player]);
                return (LRESULT) hBrush[hiscore[difficulty][gid - IDC_SCORE1].player];
            }
        acase IDC_EDIT1:
        case IDC_EDIT2:
        case IDC_EDIT3:
        case IDC_EDIT4:
        case IDC_EDIT5:
            if (hiscore[difficulty][gid - IDC_EDIT1].player != -1)
            {   // SetTextColor((HDC) wParam, 0x00000000); // black (0BGR format)
                SetBkColor((HDC) wParam, BGRColour[hiscore[difficulty][gid - IDC_EDIT1].player]);
                return (LRESULT) hBrush[hiscore[difficulty][gid - IDC_EDIT1].player];
            }
        acase IDC_TIME1:
        case IDC_TIME2:
        case IDC_TIME3:
        case IDC_TIME4:
        case IDC_TIME5:
            if (hiscore[difficulty][gid - IDC_TIME1].player != -1)
            {   SetBkColor((HDC) wParam, BGRColour[hiscore[difficulty][gid - IDC_TIME1].player]);
                return (LRESULT) hBrush[hiscore[difficulty][gid - IDC_TIME1].player];
            }
        acase IDC_DATE1:
        case IDC_DATE2:
        case IDC_DATE3:
        case IDC_DATE4:
        case IDC_DATE5:
            if (hiscore[difficulty][gid - IDC_DATE1].player != -1)
            {   SetBkColor((HDC) wParam, BGRColour[hiscore[difficulty][gid - IDC_DATE1].player]);
                return (LRESULT) hBrush[hiscore[difficulty][gid - IDC_DATE1].player];
        }   }
        return FALSE;
    default: // no need for adefault
        return FALSE;
    }
    return TRUE;
}

MODULE LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{   PAINTSTRUCT ps;
    RECT        therect;
    UINT        scancode;
    int         leftx, topy;

    /* "When a window procedure processes a message, it should return 0 from the window
       procedure. All messages that a window procedure chooses not to process must be
       passed to a Windows function named DefWindowProc(). The value returned from
       DefWindowProc() must be returned from the window procedure. */

    switch (msg)
    {
    case WM_CREATE:
     /* GetWindowRect(hwnd, &therect);
        leftx = (GetSystemMetrics(SM_CXSCREEN) / 2) - ((therect.right  - therect.left + 1) / 2);
        if (leftx < 0) leftx = 0;
        topy  = (GetSystemMetrics(SM_CYSCREEN) / 2) - ((therect.bottom - therect.top  + 1) / 2);
        if (topy < 0) topy = 0;
        DISCARD SetWindowPos
        (   hwnd,
            HWND_TOP,
            leftx,
            topy,
            0, 0,
            SWP_NOSIZE | SWP_NOCOPYBITS
        ); */

        return 0;
    case WM_SETCURSOR: // no need for acase
        if (LOWORD(lParam) == HTCLIENT)
        {   SetCursor(hPointer);
            return TRUE;
        } // implied else
        return DefWindowProc(hwnd, msg, wParam, lParam);
    case WM_PAINT: // no need for acase
        BeginPaint(hwnd, &ps);
        if (!ignorepaint)
        {   redrawscreen();
        }
        EndPaint(hwnd, &ps);
        return 0;
    case WM_KEYDOWN: // no need for acase
        if (!(lParam & 0x40000000)) // ie. if not a repeated key
        {   handleinput((SCANCODE) PARAM2SCANCODE(lParam));
        }
        return 0;
    case WM_KEYUP: // no need for acase
        scancode = PARAM2SCANCODE(lParam);
        switch (scancode)
        {
        case SCAN_LCTRL:
        case SCAN_RCTRL:
            ctrl = FALSE;
        acase SCAN_LSHIFT:
        case SCAN_RSHIFT:
            shift = FALSE;
        }
        return 0;
    case WM_CLOSE: // no need for acase
        if (verify())
        {   cleanexit(EXIT_SUCCESS);
        }
        return 0; // to avoid a spurious Visual C compiler warning
    case WM_COMMAND: // no need for acase
        switch (LOWORD(wParam))
        {
        case ID_OPTIONS_PAUSED:
            if (!paused)
            {   do_pause();
            } else
            {   globalabort = TRUE;
            }
        acase ID_FILE_UNICONIFY:
            uniconify(FALSE);
        acase ID_FILE_EXITDOS:
            cleanexit(EXIT_SUCCESS);
        }
        return 0;
    case WM_SYSCOMMAND: // no need for acase
        switch (LOWORD(wParam))
        {
        case ID_FILE_ICONIFY:
            iconify();
        acase ID_HELP_MANUAL:
            ShellExecute(0, NULL, "WORMWARS.HTM", NULL, NULL, SW_SHOW);
        acase ID_HELP_UPDATE:
            help_update();
        acase ID_HELP_ABOUT:
            DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, AboutDlgProc);
        adefault:
            return DefWindowProc(hwnd, msg, wParam, lParam);
        }
        return 0;
    case USR_NOTIFYICON: // no need for acase
        assert(iconified);
        switch (lParam)
        {
        case WM_LBUTTONDBLCLK:
            uniconify(FALSE);
        acase WM_RBUTTONUP:
            traymenu();
        }
        return 0;
    default: // no need for adefault
        return DefWindowProc(hwnd, msg, wParam, lParam);
}   }

MODULE void loadconfig(void)
{   ULONG thesize;
    FILE* LocalHandle /* = NULL */ ;

    if ((LocalHandle = fopen("WormWars.cfg", "rb")))
    {   if (fread(ConfigBuffer, 15, 1, LocalHandle) == 1)
        {   if (ConfigBuffer[0] == 9) // means V9.3+
            {   worm[0].control = ConfigBuffer[ 1];
                worm[1].control = ConfigBuffer[ 2];
                worm[2].control = ConfigBuffer[ 3];
                worm[3].control = ConfigBuffer[ 4];
                shuffle         = ConfigBuffer[ 5];
                engraved        = ConfigBuffer[ 6];
                if
                (   entirex >= resolutions[ConfigBuffer[7]].width
                 && entirey >= resolutions[ConfigBuffer[7]].height
                )
                {   resolution  = ConfigBuffer[ 7];
                }
                soundfx         = ConfigBuffer[ 8];
                music           = ConfigBuffer[ 9];
                difficulty      = ConfigBuffer[10];
                language        = ConfigBuffer[11];
                isometric       = ConfigBuffer[12];
                autosave        = ConfigBuffer[13];
                tdworms         = ConfigBuffer[14];
        }   }
        DISCARD fclose(LocalHandle);
        // LocalHandle = NULL;
}   }

MODULE void saveconfig(void)
{   FILE* LocalHandle /* = NULL */ ;

    if ((LocalHandle = fopen("WormWars.cfg", "wb")))
    {   ConfigBuffer[0]  = (UBYTE) 9; // means V9.3+
        ConfigBuffer[1]  = worm[0].control;
        ConfigBuffer[2]  = worm[1].control;
        ConfigBuffer[3]  = worm[2].control;
        ConfigBuffer[4]  = worm[3].control;
        ConfigBuffer[5]  = shuffle;
        ConfigBuffer[6]  = engraved;
        ConfigBuffer[7]  = resolution;
        ConfigBuffer[8]  = soundfx;
        ConfigBuffer[9]  = music;
        ConfigBuffer[10] = difficulty;
        ConfigBuffer[11] = (UBYTE) language;
        ConfigBuffer[12] = isometric;
        ConfigBuffer[13] = autosave;
        ConfigBuffer[14] = tdworms;

        DISCARD fwrite(ConfigBuffer, 15, 1, LocalHandle);
        DISCARD fclose(LocalHandle);
        // LocalHandle = NULL;
}   }

MODULE void killfocus(HWND hwnd, int focussed)
{   effect(FXAPPLAUSE);
    GetWindowText(GetDlgItem(hwnd, IDC_EDIT1 + focussed), worm[hiscore[difficulty][focussed].player].name, NAMELENGTH);
    if (worm[hiscore[difficulty][focussed].player].name[0] >= 'a' && worm[hiscore[difficulty][focussed].player].name[0] <= 'z')
    {   worm[hiscore[difficulty][focussed].player].name[0] -= 32;
    }
    strcpy(hiscore[difficulty][focussed].name, worm[hiscore[difficulty][focussed].player].name);
    SetWindowText(GetDlgItem(hwnd, IDC_EDIT1 + focussed), worm[hiscore[difficulty][focussed].player].name);
    SendMessage(GetDlgItem(hwnd, IDC_EDIT1 + focussed), EM_SETREADONLY, TRUE, 0);
    hiscore[difficulty][focussed].fresh = FALSE;
    next(hwnd);
}

EXPORT void intro(void)
{   HBITMAP OldBitmapPtr;
    HFONT   OldFontPtr;
    TEXT    tempstring[80 + 1];
    int     cornerx, cornery,
            i,
            x, y;

    showing = 5;

    if (level >= 1 && level <= TUTORIALS)
    {   if (language == LANGUAGE_ENG)
        {   rectfill_black(xoffset + 14, yoffset - 190, xoffset + 427, yoffset + 113);
        } else
        {   rectfill_black(xoffset - 46, yoffset - 190, xoffset + 487, yoffset + 113);
    }   }
    else
    {   if (language == LANGUAGE_ENG)
        {   rectfill_black(xoffset + 14, yoffset -  10, xoffset + 427, yoffset + 113);
        } else
        {   rectfill_black(xoffset - 46, yoffset -  10, xoffset + 487, yoffset + 113);
    }   }

    cornerx = xoffset - 20;
    cornery = yoffset - 18;

    for (x = ((language == LANGUAGE_ENG) ? 2 : -3); x <= ((language == LANGUAGE_ENG) ? 37 : 42); x++)
    {   if (level >= 1 && level <= TUTORIALS)
        {   raw_draw(cornerx + ( x * SQUAREX), cornery - (15 * SQUAREY), FROST);
        }
        raw_draw(cornerx + ( x * SQUAREX), cornery                 , FROST);
        raw_draw(cornerx + ( x * SQUAREX), cornery + (11 * SQUAREY), FROST);
    }
    for (y = ((level >= 1 && level <= TUTORIALS) ? (-15) : 0); y <= 11; y++)
    {   raw_draw(cornerx + (((language == LANGUAGE_ENG) ?  2 : -3) * SQUAREX), cornery + ( y * SQUAREY), FROST);
        raw_draw(cornerx + (((language == LANGUAGE_ENG) ? 37 : 42) * SQUAREX), cornery + ( y * SQUAREY), FROST);
    }

    for (i = 0; i <= SPECIES; i++)
    {   if (creatureinfo[i].introlevel == level)
        {   raw_draw(xoffset + INTROTILEX, yoffset + (SQUAREY * 3), creatureinfo[i].symbol);
            break; // for speed
    }   }
    for (i = 0; i <= LASTOBJECT; i++)
    {   if (objectinfo[i].introlevel == level)
        {   raw_draw(xoffset + INTROTILEX, yoffset + (SQUAREY * 5), (UWORD) i);
            break; // for speed
    }   }
    if (level < LASTFRUIT - FIRSTFRUIT + 1)
    {   raw_draw(    xoffset + INTROTILEX, yoffset + (SQUAREY * 7), (UWORD) (level + FIRSTFRUIT - 1));
    }

    OnScreenRastPort = GetDC(MainWindowPtr);
    OldFontPtr = SelectObject(OnScreenRastPort, hBigFont);
    SetTextColor(OnScreenRastPort, WHITE);
    SetBkColor(OnScreenRastPort, BLACK);

    if (level >= 1 && level <= TUTORIALS)
    {   dotext(    xoffset + RUNDOWNX_1ST, yoffset - 4 - (SQUAREY * 14), getintrotext());
    }
    dotext(        xoffset + RUNDOWNX_1ST, yoffset - 4 + (SQUAREY *  1), (STRPTR) GetCatalogStr(CatalogPtr, MSG_INTRODUCING,   "Introducing:"));
    for (i = 0; i <= SPECIES; i++)
    {   if (creatureinfo[i].introlevel == level)
        {   dotext(xoffset + INTROTEXTX,   yoffset - 2 + (SQUAREY *  3), creatureinfo[i].name);
            break; // for speed
    }   }
    for (i = 0; i <= LASTOBJECT; i++)
    {   if (objectinfo[i].introlevel == level)
        {   dotext(xoffset + INTROTEXTX,   yoffset - 2 + (SQUAREY *  5), objectinfo[i].desc);
            break; // for speed
    }   }
    if (level < LASTFRUIT - FIRSTFRUIT + 1)
    {   sprintf
        (   tempstring,
            "%s: %d %s.",
            fruitname[level - 1],
            POINTS_FRUIT * level,
            GetCatalogStr(CatalogPtr, MSG_POINTS2, "points")
        );
        dotext(    xoffset + INTROTEXTX,   yoffset - 2 + (SQUAREY *  7), tempstring);
    }

    SelectObject(OnScreenRastPort, OldFontPtr);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;

    saylevel(WHITE);
}

EXPORT void dotext(int x, int y, STRPTR thestring)
{   assert(OnScreenRastPort);
    TextOut(OnScreenRastPort, x, y, thestring, strlen(thestring));
}

MODULE void renderhiscores(HWND hwnd)
{   TRANSIENT int  i;
    PERSIST   TEXT OutputString[80 + 1];

    for (i = 0; i <= HISCORES; i++)
    {   SendMessage(GetDlgItem(hwnd, IDC_EDIT1 + i), EM_LIMITTEXT, (WPARAM) NAMELENGTH, 0);
        if (hiscore[difficulty][i].player == -1)
        {   OutputString[0] = EOS;
            SetDlgItemText(hwnd, IDC_EDIT1 + i, OutputString);
        } else
        {   SetDlgItemText(hwnd, IDC_EDIT1 + i, worm[hiscore[difficulty][i].player].name);
            sprintf(OutputString, "%d", hiscore[difficulty][i].score);
        }
        SetDlgItemText(hwnd, IDC_SCORE1 + i, OutputString);

        if (hiscore[difficulty][i].fresh)
        {   SendMessage(GetDlgItem(hwnd, IDC_EDIT1 + i), EM_SETREADONLY, FALSE, 0);
        }

        if (hiscore[difficulty][i].level == -1)
        {   strcpy(OutputString, (STRPTR) GetCatalogStr(CatalogPtr, MSG_ALL, "All"));
        } elif (hiscore[difficulty][i].level == 0)
        {   OutputString[0] = EOS;
        } else
        {   sprintf(OutputString, "%d", hiscore[difficulty][i].level);
        }
        SetDlgItemText(hwnd, IDC_LEVEL1 + i, OutputString);

        if (hiscore[difficulty][i].time[0])
        {   strcpy(OutputString, hiscore[difficulty][i].time);
        } else
        {   OutputString[0] = EOS;
        }
        SetDlgItemText(hwnd, IDC_TIME1 + i, OutputString);

        if (hiscore[difficulty][i].date[0])
        {   strcpy(OutputString, hiscore[difficulty][i].date);
        } else
        {   OutputString[0] = EOS;
        }
        SetDlgItemText(hwnd, IDC_DATE1 + i, OutputString);
}   }

MODULE void do_pause(void)
{   int  i;
    FLAG rc;

    say((STRPTR) GetCatalogStr(CatalogPtr, MSG_PAUSED, "Paused...press P to continue"), WHITE);
    paused = TRUE;
    if (TrayMenuPtr)
    {   DISCARD CheckMenuItem(TrayMenuPtr, (UINT) ID_OPTIONS_PAUSED, MF_BYCOMMAND | MF_CHECKED);
    }
    rc = busyloop(FALSE);
    paused = FALSE;
    if (TrayMenuPtr)
    {   DISCARD CheckMenuItem(TrayMenuPtr, (UINT) ID_OPTIONS_PAUSED, MF_BYCOMMAND | MF_UNCHECKED);
    }
    say((STRPTR) GetCatalogStr(CatalogPtr, MSG_UNPAUSED, "Unpaused"), WHITE);
    if (secondsleft <= 0)
    {   say((STRPTR) GetCatalogStr(CatalogPtr, MSG_TIMEEXPIRED, "Time expired!"), WHITE);
        secondsleft = 0;
    }
    if (!rc)
    {   aborted = TRUE;
        worm[0].lives = worm[1].lives = worm[2].lives = worm[3].lives = 0;
        if (level)
        {   for (i = 0; i <= 3; i++)
            {   worm[i].levelreached = level;
        }   }
        else
        {   for (i = 0; i <= 3; i++)
            {   worm[i].levelreached = reallevel;
}   }   }   }

EXPORT void fileopen(FLAG revert)
{   PERSIST   TEXT         newpathname[MAX_PATH + 1],
                           OutputStr[MAX_PATH + 40],
                           tempstring[MAX_PATH + 80];
    TRANSIENT OPENFILENAME ofn;

    if
    (   (revert && pathname[0] == EOS)
     || ((!revert || !autosave) && !verify())
    )
    {   return;
    }

    strcpy(newpathname, pathname);

    if (!revert)
    {   ZeroMemory(&ofn, sizeof(ofn));
        ofn.lStructSize     = sizeof(ofn);
        ofn.hwndOwner       = (a == LEVELEDIT) ? MainWindowPtr : TitleWindowPtr;
        ofn.lpstrTitle      = (STRPTR) GetCatalogStr(CatalogPtr, MSG_OPENLSET, "Open Levelset"); // we should have a separate hail string
        ofn.lpstrFilter     = "Levelset Files (*.LSET)\0*.LSET\0" \
                              "All Files (*.*)\0*.*\0"; // we should localize this
        ofn.lpstrFile       = newpathname;
        ofn.nMaxFile        = MAX_PATH;
        ofn.Flags           = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;

        if (!GetOpenFileName(&ofn))
        {   return;
    }   }

    if (!loadfields(newpathname))
    {   strcpy(pathname, newpathname);
        if (a == LEVELEDIT)
        {   if (level > levels)
            {   level = levels;
            }
            sprintf
            (   tempstring,
                GetCatalogStr(CatalogPtr, MSG_OPENED, "Opened %s (%d levels)."),
                newpathname,
                levels
            );
            say(tempstring, WHITE);
            le_drawfield();
        } else
        {   assert(a == TITLESCREEN);
            renderhiscores(TitleWindowPtr);
            EnableWindow(GetDlgItem(TitleWindowPtr, IDC_CLEARHISCORES), hiscoresareclear() ? FALSE : TRUE);
    }   }
    else
    {   sprintf
        (   OutputStr,
            (STRPTR) GetCatalogStr
            (   CatalogPtr,
                MSG_CANTOPEN,
                "Can't open %s!"
            ),
            newpathname
        );
        MessageBox(NULL, OutputStr, "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
}   }

EXPORT void filesaveas(FLAG saveas)
{   PERSIST   TEXT         newpathname[MAX_PATH + 1],
                           OutputStr[MAX_PATH + 40],
                           tempstring[MAX_PATH + 80];
    TRANSIENT OPENFILENAME ofn;

    strcpy(newpathname, pathname);

    if (saveas || newpathname[0] == EOS)
    {   if
        (   strlen(newpathname) < 5
         || stricmp(&newpathname[strlen(newpathname) - 5], ".lset")
        )
        {   strcat(newpathname, ".lset");
        }

        ZeroMemory(&ofn, sizeof(ofn));
        ofn.lStructSize     = sizeof(ofn);
        ofn.hwndOwner       = (a == LEVELEDIT) ? MainWindowPtr : TitleWindowPtr;
        ofn.lpstrTitle      = (STRPTR) GetCatalogStr(CatalogPtr, MSG_SAVELSET, "Save Levelset"); // we should have a separate hail string
        ofn.lpstrFilter     = "Levelset Files (*.LSET)\0*.LSET\0" \
                              "All Files (*.*)\0*.*\0"; // we should localize this
        ofn.lpstrFile       = newpathname;
        ofn.nMaxFile        = MAX_PATH;
        ofn.Flags           = OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;

        if (!GetSaveFileName(&ofn))
        {   return;
        }

        if
        (   strlen(newpathname) < 5
         || stricmp(&newpathname[strlen(newpathname) - 5], ".lset")
        )
        {   strcat(newpathname, ".lset");
    }   }

    if (savefields(newpathname))
    {   strcpy(pathname, newpathname);
        levels_modified = scores_modified = FALSE;

        if (a == LEVELEDIT)
        {   sprintf
            (   tempstring,
                GetCatalogStr(CatalogPtr, MSG_SAVED,    "Saved %s (%d levels)."),
                pathname,
                levels
            );
            say(tempstring, WHITE);
    }   }
    else
    {   sprintf
        (   OutputStr,
            (STRPTR) GetCatalogStr
            (   CatalogPtr,
                MSG_CANTSAVE,
                "Can't save %s!"
            ),
            newpathname
        );
        MessageBox(NULL, OutputStr, "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
}   }

/* Some relevant locale IDs are:
   $4 $07 = German
   $4 $09 = American
   $4 $0C = French
   $4 $10 = Italian
   $8 $09 = English
   $C $09 = Australian
   $? $13 = Dutch
   $? $15 = Polish
   $? $19 = Russian */

MODULE void freelanguage(void)
{   if (langs[language].winbuffer)
    {   free(langs[language].winbuffer);
        langs[language].winbuffer = NULL;
    }
    if (langs[language].offset)
    {   free(langs[language].offset);
        langs[language].offset = NULL;
}   }

MODULE void load_catalog(void)
{   UBYTE        *p8data,
                 *fbase;
    UBYTE         LocalIOBuffer[8];
    UBYTE*        dataptr = NULL;
    int           code,
                  i;
    struct Chunk* p8Chunk;
    ULONG         fsize,
                  thesize;
    FILE*         TheLocalHandle; // LocalHandle is a variable of winbase.h

/* Open the file.
   Read the first 8 bytes.
   Assure it is a FORM.
   Allocate enough memory to read the rest of the FORM chunk.
   Read the rest of the FORM chunk.
   Close the file. */

 // groupsep = langs[language].groupsep;
 // decpoint = langs[language].decpoint;

    if (language == LANGUAGE_ENG)
    {   return;
    }

    if (!(TheLocalHandle = fopen(langs[language].filename, "rb")))
    {   code = 1;                               /* can't open file */
        goto DONE;
    }
    if (fread(LocalIOBuffer, 8, 1, TheLocalHandle) != 1)
    {   code = 2;                               /* can't read file (2), or too short (3) */
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }

    p8Chunk = (struct Chunk *) &LocalIOBuffer[0];
    if (p8Chunk->ckID != (ULONG) ID_FORM)
    {   code = 3;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    fsize = (LocalIOBuffer[4] * 16777216)
          + (LocalIOBuffer[5] *    65536)
          + (LocalIOBuffer[6] *      256)
          +  LocalIOBuffer[7];
    if (!(fbase = (UBYTE *) malloc(fsize)))
    {   code = 4;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    p8data = fbase;
    if (fread(p8data, fsize, 1, TheLocalHandle) != 1)
    {   code = 5;
        DISCARD fclose(TheLocalHandle);
        // TheLocalHandle = NULL;
        goto DONE;
    }
    DISCARD fclose(TheLocalHandle);
    // TheLocalHandle = NULL;

/* Assure it is a CTLG.
   Look at each chunk: if it is a STRS, remember offset and length.
   Allocate enough memory to store a copy of the STRS chunk.
   Copy the STRS chunk. */

    if (MAKE_ID(*p8data, *(p8data + 1), *(p8data + 2), *(p8data + 3)) != ID_CTLG)
    {   code = 6;
        goto DONE;
    }
    p8data += 4;

    while (p8data < fbase + fsize)
    {   thesize = (p8data[4] * 16777216)
                + (p8data[5] *    65536)
                + (p8data[6] *      256)
                +  p8data[7];
        p8Chunk = (struct Chunk *) p8data;
        switch (p8Chunk->ckID)
        {
        case ID_STRS:
            langs[language].size = (int) thesize;
            dataptr = (UBYTE *) (p8data + 8);
        }
        p8data += 8 + thesize;
        if ((thesize & 1) == 1)
        {   p8data++;
    }   }

    if (dataptr == NULL) // no STRS chunk found
    {   code = 7;
        goto DONE;
    }

    langs[language].winbuffer = (UBYTE *) malloc((ULONG) langs[language].size);
    if (langs[language].winbuffer == NULL)
    {   code = 8;
        goto DONE;
    }

    code = 0; // success
    memcpy(langs[language].winbuffer, dataptr, (unsigned int) langs[language].size);
    switch (language)
    {
    case LANGUAGE_GRE:
        // convert from ELOT 928 to Windows-1253
        for (i = 0; i < langs[language].size; i++)
        {   if     (langs[language].winbuffer[i] == 181)
            {   langs[language].winbuffer[i] = 161;
            } elif (langs[language].winbuffer[i] == 182)
            {   langs[language].winbuffer[i] = 162;
        }   }
    acase LANGUAGE_POL:
        // convert from ISO-8859-2 to Windows-1250
        for (i = 0; i < langs[language].size; i++)
        {   switch (langs[language].winbuffer[i])
            {
            case  161: langs[language].winbuffer[i] = 165; // 
            acase 165: langs[language].winbuffer[i] = 188; // 
            acase 166: langs[language].winbuffer[i] = 140; // 
            acase 169: langs[language].winbuffer[i] = 138; // 
            acase 171: langs[language].winbuffer[i] = 141; // 
            acase 172: langs[language].winbuffer[i] = 143; // 
            acase 174: langs[language].winbuffer[i] = 142; // 
            acase 177: langs[language].winbuffer[i] = 185; // 
            acase 181: langs[language].winbuffer[i] = 190; // 
            acase 182: langs[language].winbuffer[i] = 156; // 
            acase 183: langs[language].winbuffer[i] = 161; // 
            acase 185: langs[language].winbuffer[i] = 154; // 
            acase 187: langs[language].winbuffer[i] = 157; // 
            acase 188: langs[language].winbuffer[i] = 159; // 
            acase 190: langs[language].winbuffer[i] = 158; // 
    }   }   }

    free(fbase);
    // fbase = NULL;

DONE:
    if (code)
    {   // OC; printf("Error %d!\n", code);
        MessageBox(NULL, "Can't load catalog file!", "Worm Wars: Error", MB_ICONEXCLAMATION | MB_OK);
    } else
    {   // OC; printf("Loaded \"%s\" (%d bytes at $%X).\n\n", langs[language].filename, langs[language].size, dataptr);
        parse_catalog();
}   }

EXPORT STRPTR GetCatalogStr(int catalog, LONG stringNum, STRPTR defaultString)
{   if
    (   language                          == LANGUAGE_ENG
     || langs[language].winbuffer         == NULL
     || langs[language].offset            == NULL
     || stringNum                         >  highestid
     || langs[language].offset[stringNum] == 0
    )
    {   return defaultString;
    } // implied else
    return &langs[language].winbuffer[langs[language].offset[stringNum]];
}

MODULE void parse_catalog(void)
{   int i,
        id,
        length;

    assert(langs[language].winbuffer);

    // find the highest ID
    i = 0;
    highestid = 0;
    while (i < langs[language].size)
    {   id     = (langs[language].winbuffer[i    ] * 16777216)
               + (langs[language].winbuffer[i + 1] *    65536)
               + (langs[language].winbuffer[i + 2] *      256)
               + (langs[language].winbuffer[i + 3]);
        if (id > highestid)
        {   highestid = id;
        }
        length = (langs[language].winbuffer[i + 4] * 16777216)
               + (langs[language].winbuffer[i + 5] *    65536)
               + (langs[language].winbuffer[i + 6] *      256)
               + (langs[language].winbuffer[i + 7]);

        i += 8 + length + 1;
        if   ((i % 4) == 1) i += 3;
        elif ((i % 4) == 2) i += 2;
        elif ((i % 4) == 3) i++;
    }

    langs[language].offset = (ULONG*) calloc((size_t) (4 * (highestid + 1)), 1);

    // fill langs[].offset[] array
    i = 0;
    while (i < langs[language].size)
    {   id     = (langs[language].winbuffer[i    ] * 16777216)
               + (langs[language].winbuffer[i + 1] *    65536)
               + (langs[language].winbuffer[i + 2] *      256)
               + (langs[language].winbuffer[i + 3]);
        length = (langs[language].winbuffer[i + 4] * 16777216)
               + (langs[language].winbuffer[i + 5] *    65536)
               + (langs[language].winbuffer[i + 6] *      256)
               + (langs[language].winbuffer[i + 7]);
        langs[language].offset[id] = (ULONG) (i + 8);
        // OC; printf("#%d: %s\n", id, &langs[language].winbuffer[langs[language].offset[id]]);

        i += 8 + length + 1;
        if   ((i % 4) == 1) i += 3;
        elif ((i % 4) == 2) i += 2;
        elif ((i % 4) == 3) i++;
}   }

MODULE void retranslate(void)
{   if (TitleWindowPtr)
    {   setdlgtext(TitleWindowPtr, IDC_LOAD,          MSG_OPENLSET2,    "&Load levelset...");
        setdlgtext(TitleWindowPtr, IDC_SAVE,          MSG_SAVELSET2,    "&Save levelset...");
        setdlgtext(TitleWindowPtr, IDC_UPDATE,        MSG_MENU_UPDATE,  "Check for updates...");
        setdlgtext(TitleWindowPtr, IDC_CLEARHISCORES, MSG_MENU_CLEARHS, "Clear high scores");
        setdlgtext(TitleWindowPtr, IDC_3DWORMS,       MSG_3DWORMS,      "&3D worms?");
        setdlgtext(TitleWindowPtr, IDC_AUTOSAVE,      MSG_AUTOSAVE,     "Autosave levelset?");
        setdlgtext(TitleWindowPtr, IDC_ENGRAVED,      MSG_ENGRAVED2,    "Engraved s&quares?");
        setdlgtext(TitleWindowPtr, IDC_ISOMETRIC,     MSG_ISOMETRIC,    "&Isometric?");
        setdlgtext(TitleWindowPtr, IDC_SHUFFLE,       MSG_SHUFFLE2,     "Shuffle le&vels?");
        setdlgtext(TitleWindowPtr, IDC_EASY,          MSG_EASY,         "Easy");
        setdlgtext(TitleWindowPtr, IDC_NORMAL,        MSG_NORMAL,       "Normal");
        setdlgtext(TitleWindowPtr, IDC_HARD,          MSG_HARD,         "Hard");
        setdlgtext(TitleWindowPtr, IDC_VERYHARD,      MSG_VERYHARD,     "Very Hard");
        setdlgtext(TitleWindowPtr, IDC_EFFECTS,       MSG_FX2,          "E&ffects?");
        setdlgtext(TitleWindowPtr, IDC_MUSIC,         MSG_MUSIC2,       "&Music?");
        setdlgtext(TitleWindowPtr, IDC_ENG,           MSG_ENG,          "English");
        setdlgtext(TitleWindowPtr, IDC_GER,           MSG_GER,          "German");
        setdlgtext(TitleWindowPtr, IDC_GRE,           MSG_GRE,          "Greek");
        setdlgtext(TitleWindowPtr, IDC_ITA,           MSG_ITA,          "Italian");
        setdlgtext(TitleWindowPtr, IDC_POL,           MSG_POL,          "Polish");
        setdlgtext(TitleWindowPtr, IDC_SPA,           MSG_SPA,          "Spanish");
        setdlgtext(TitleWindowPtr, IDC_START,         MSG_STARTGAME2,   "Start &Game");
        setdlgtext(TitleWindowPtr, IDC_LEVELEDITOR,   MSG_LEVELEDITOR2, "Level &Editor");
        setdlgtext(TitleWindowPtr, IDC_CREATURES,     MSG_CREATURES,    "&Creatures...");
        setdlgtext(TitleWindowPtr, IDC_OBJECTS,       MSG_OBJECTS,      "&Objects...");
        setdlgtext(TitleWindowPtr, IDC_FRUITS,        MSG_FRUITS,       "F&ruits...");
        setdlgtext(TitleWindowPtr, IDC_MANUAL,        MSG_MANUAL,       "Ma&nual...");
        setdlgtext(TitleWindowPtr, IDC_ABOUT,         MSG_ABOUT,        "&About...");
        setdlgtext(TitleWindowPtr, IDC_GREEN_NONE,    MSG_NONE,         "None");
        setdlgtext(TitleWindowPtr, IDC_RED_NONE,      MSG_NONE,         "None");
        setdlgtext(TitleWindowPtr, IDC_BLUE_NONE,     MSG_NONE,         "None");
        setdlgtext(TitleWindowPtr, IDC_YELLOW_NONE,   MSG_NONE,         "None");
        setdlgtext(TitleWindowPtr, IDC_GREEN_KYBD,    MSG_LEFTKYBD,     "Lt. Kybd");
        setdlgtext(TitleWindowPtr, IDC_RED_KYBD,      MSG_RIGHTKYBD,    "Rt. Kybd");
        setdlgtext(TitleWindowPtr, IDC_BLUE_JOY,      MSG_JOYSTICK1,    "Joystick 1");
        setdlgtext(TitleWindowPtr, IDC_YELLOW_JOY,    MSG_JOYSTICK2,    "Joystick 2");

        setdlgtext(TitleWindowPtr, IDL_HIGHSCORES,    MSG_HIGHSCORES,   "High Score Table");
        setdlgtext(TitleWindowPtr, IDL_WORMCONTROLS,  MSG_WORMCONTROLS, "Worm Controls");
        setdlgtext(TitleWindowPtr, IDL_GREENWORM,     MSG_GREENWORM,    "Green Worm");
        setdlgtext(TitleWindowPtr, IDL_REDWORM,       MSG_REDWORM,      "Red Worm");
        setdlgtext(TitleWindowPtr, IDL_BLUEWORM,      MSG_BLUEWORM,     "Blue Worm");
        setdlgtext(TitleWindowPtr, IDL_YELLOWWORM,    MSG_YELLOWWORM,   "Yellow Worm");
        setdlgtext(TitleWindowPtr, IDL_RESOLUTION,    MSG_RESOLUTION,   "Resolution");
        setdlgtext(TitleWindowPtr, IDL_DIFFICULTY,    MSG_DIFFICULTY2,  "Difficulty");
        setdlgtext(TitleWindowPtr, IDL_SOUND,         MSG_SOUND,        "Sound");
        setdlgtext(TitleWindowPtr, IDL_LANGUAGE,      MSG_LANGUAGE,     "Language");
        setdlgtext(TitleWindowPtr, IDL_OPTIONS,       MSG_OPTIONS,      "Options");
        setdlgtext(TitleWindowPtr, IDL_SCORE,         MSG_SCORE,        "Score");
        setdlgtext(TitleWindowPtr, IDL_LEVEL,         MSG_LEVEL2,       "Level");
        setdlgtext(TitleWindowPtr, IDL_NAME,          MSG_NAME,         "Name");
        setdlgtext(TitleWindowPtr, IDL_TIME,          MSG_TIME,         "Time");
        setdlgtext(TitleWindowPtr, IDL_DATE,          MSG_DATE,         "Date");
     // setdlgtext(TitleWindowPtr, IDL_HELP,          MSG_MENU_HELP2,   "Help");
}   }

EXPORT void introanim(void)
{   FLAG        done;
    ULONG       oldtime;
    MSG         Message;
    UINT        scancode;
    PAINTSTRUCT ps;

    if (!MainWindowPtr)
    {   return;
    }

    assert(level <= TUTORIALS);

    introframe = 0;
    for (;;)
    {   introanim_engine(introframe);
        if (introframe <= 31)
        {   introframe++;
        }

        done = FALSE;
        oldtime = timeGetTime();
        do
        {   Sleep(0);
            while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
            {   switch (Message.message)
                {
                case WM_PAINT:
                    if (ignorepaint)
                    {   BeginPaint(MainWindowPtr, &ps);
                        EndPaint(MainWindowPtr, &ps);
                        ignorepaint = FALSE;
                    } else
                    {   BeginPaint(MainWindowPtr, &ps);
                        redrawscreen();
                        EndPaint(MainWindowPtr, &ps);
                    }
                acase WM_LBUTTONDOWN:
                    done = TRUE;
                acase WM_KEYDOWN:
                    if (!(Message.lParam & 0x40000000)) // ie. if not a repeated key
                    {   scancode = PARAM2SCANCODE(Message.lParam);
                        switch (scancode)
                        {
                        case SCAN_ESCAPE:
                            done = TRUE;
                        acase SCAN_F:
                        case SCAN_M:
                            toggle((SBYTE) scancode);
                            done = TRUE;
                        acase SCAN_F12:
                            if (ctrl)
                            {   iconify();
                            }
                        acase SCAN_LCTRL:
                        case SCAN_RCTRL:
                            ctrl = TRUE;
                        acase SCAN_LSHIFT:
                        case SCAN_RSHIFT:
                            shift = TRUE;
                        adefault:
                            done = TRUE;
                    }   }
                acase WM_KEYUP:
                    scancode = PARAM2SCANCODE(Message.lParam);
                    switch (scancode)
                    {
                    case SCAN_LCTRL:
                    case SCAN_RCTRL:
                        ctrl = FALSE;
                    acase SCAN_LSHIFT:
                    case SCAN_RSHIFT:
                        shift = FALSE;
                    }
                adefault:
                    TranslateMessage(&Message);
                    DispatchMessage(&Message);
            }   }
            if (firebutton())
            {   done = TRUE;
        }   }
        while (!done && timeGetTime() < oldtime + INTROSPEED);

        if (done)
        {   return;
}   }   }

EXPORT void iconify(void)
{   NOTIFYICONDATA tnid;
    BOOL           pvParam;
    int            i;

    if (iconified)
    {   return;
    }

    tnid.cbSize           = sizeof(NOTIFYICONDATA);
    tnid.hWnd             = CurrentWindowPtr;
    tnid.uID              = 1;
    tnid.uFlags           = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    tnid.uCallbackMessage = USR_NOTIFYICON;
    tnid.hIcon            = (HICON) LoadImage(InstancePtr, MAKEINTRESOURCE(IDI_WORMWARS), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
    strcpy(tnid.szTip, "Worm Wars");

    DISCARD Shell_NotifyIcon(NIM_ADD, &tnid);

    ShowWindow(CurrentWindowPtr,  SW_HIDE);

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == FALSE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, TRUE, 0, SPIF_SENDWININICHANGE);
    }   }

    iconified = TRUE;
}

MODULE void uniconify(FLAG exiting)
{   NOTIFYICONDATA tnid;
    BOOL           pvParam;
    int            i;

    assert(iconified);

    if (!exiting)
    {   ShowWindow(CurrentWindowPtr, SW_SHOW);
        SetActiveWindow(CurrentWindowPtr);
    }

    tnid.cbSize = sizeof(NOTIFYICONDATA);
    tnid.hWnd   = CurrentWindowPtr;
    tnid.uID    = 1;
    DISCARD Shell_NotifyIcon(NIM_DELETE, &tnid);
    DestroyMenu(TrayMenuPtr);
    TrayMenuPtr = NULL;

    if (lame)
    {   SystemParametersInfo(SPI_GETSCREENSAVEACTIVE, 0, &pvParam, 0);
        if (pvParam == TRUE)
        {   SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, FALSE, 0, SPIF_SENDWININICHANGE);
    }   }

    ctrl = shift = FALSE;
    iconified = FALSE;
}

MODULE void traymenu(void)
{   int          i;
    HMENU        TraySubMenuPtr;
    POINT        ThePoint;
    MENUITEMINFO mii =
    {   44,
        MIIM_STATE,
        0,
        MFS_DEFAULT, // to embolden it
        ID_FILE_ICONIFY,
        NULL,
        NULL,
        NULL,
        ID_FILE_ICONIFY,
        0,
        0
    };

    assert(iconified);

    DISCARD SetForegroundWindow(CurrentWindowPtr);
    DISCARD GetCursorPos(&ThePoint);

    TrayMenuPtr = LoadMenu(InstancePtr, MAKEINTRESOURCE(IDR_TRAYMENU));
    for (i = 0; i < MENUBITMAPS; i++)
    {   DISCARD SetMenuItemBitmaps
        (   TrayMenuPtr,     // handle of menu
            mib[i].menuitem, // menu item to receive new bitmaps
            MF_BYCOMMAND,    // menu item flags
            mib[i].bitmap,   // handle of unchecked bitmap
            NULL             // handle of checked bitmap
        );
    }
    TraySubMenuPtr = GetSubMenu(TrayMenuPtr, 0);
    DISCARD SetMenuItemInfo
    (   TraySubMenuPtr,
        ID_FILE_UNICONIFY,
        FALSE,
        &mii
    );
    if (a == PLAYGAME)
    {   DISCARD EnableMenuItem(TrayMenuPtr, (UINT) ID_OPTIONS_PAUSED, MF_BYCOMMAND | MF_ENABLED);
    } else
    {   DISCARD EnableMenuItem(TrayMenuPtr, (UINT) ID_OPTIONS_PAUSED, MF_BYCOMMAND | MF_GRAYED);
    }
    DISCARD CheckMenuItem(TrayMenuPtr, (UINT) ID_OPTIONS_PAUSED, MF_BYCOMMAND | (paused ? MF_CHECKED : MF_UNCHECKED));
    TrackPopupMenu
    (   TraySubMenuPtr,
        TPM_RIGHTALIGN | TPM_RIGHTBUTTON,
        ThePoint.x,
        ThePoint.y,
        0,
        CurrentWindowPtr,
        NULL
    );
    PostMessage(CurrentWindowPtr, WM_NULL, 0, 0);
}

MODULE void repaint_intro(void)
{   int i;

    for (i = 0; i <= introframe; i++)
    {   introanim_engine(i);
}   }

MODULE void setdlgtext(HWND hwnd, int gid, int msg, STRPTR defaultmsg)
{   switch (language)
    {
    case LANGUAGE_GRE:
        DISCARD SendMessage(GetDlgItem(hwnd, gid), WM_SETFONT, (WPARAM) hGreekFont , MAKELPARAM(FALSE, 0));
    acase LANGUAGE_POL:
        DISCARD SendMessage(GetDlgItem(hwnd, gid), WM_SETFONT, (WPARAM) hPolishFont, MAKELPARAM(FALSE, 0)); // this seems unnecessary, on Win8.1 at least
    }
    DISCARD SetDlgItemText(hwnd, gid, GetCatalogStr(CatalogPtr, msg, defaultmsg));
}

EXPORT void openconsole(void)
{   int   hScrn, hKybd;
    FILE* hf;

    if (consoleopen)
    {   return;
    }

    assert(!consoleopen);

    AllocConsole(); // might need freeing at program exit
    consoleopen = TRUE;
    hScrn = _open_osfhandle((long) GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT); // might need freeing at program exit
    hKybd = _open_osfhandle((long) GetStdHandle(STD_INPUT_HANDLE ), _O_TEXT); // might need freeing at program exit
    hf = _fdopen(hScrn, "w"); // might need freeing at program exit
    *stdout = *hf;
    hf = _fdopen(hKybd, "r"); // might need freeing at program exit
    *stdin = *hf;
    DISCARD setvbuf(stdout, NULL, _IONBF, 0);
    DISCARD setvbuf(stdin , NULL, _IONBF, 0);

    DISCARD SetConsoleTitle("Worm Wars Console");
}

EXPORT void writetext(int x, int y, COLOUR colour, STRPTR text)
{   HFONT OldFontPtr;
    HDC   OnScreenRastPort = NULL;

    if (!MainWindowPtr)
    {   return;
    }

    y -= 9;

    OnScreenRastPort = GetDC(MainWindowPtr);
    OldFontPtr = SelectObject(OnScreenRastPort, hBoldFont);
    SetBkMode(OnScreenRastPort, TRANSPARENT);
    SetTextColor(OnScreenRastPort, BLACK);
    TextOut(OnScreenRastPort, x + 1, y + 1, text, strlen(text));
    SetTextColor(OnScreenRastPort, colour);
    TextOut(OnScreenRastPort, x, y, text, strlen(text));
    SelectObject(OnScreenRastPort, OldFontPtr);
    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void rectfill_black(int leftx, int topy, int rightx, int bottomy)
{   RECT therect;
    HDC  OnScreenRastPort = NULL;

    assert(MainWindowPtr);

    therect.left   = leftx;
    therect.top    = topy;
    therect.right  = rightx + 1;
    therect.bottom = bottomy + 1;
    OnScreenRastPort = GetDC(MainWindowPtr);

    FillRect
    (   OnScreenRastPort,
        &therect,
        GetStockObject(BLACK_BRUSH)
    );

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}
EXPORT void rectfill_grey(int leftx, int topy, int rightx, int bottomy)
{   RECT therect;
    HDC  OnScreenRastPort = NULL;

    assert(MainWindowPtr);

    therect.left   = leftx;
    therect.top    = topy;
    therect.right  = rightx + 1;
    therect.bottom = bottomy + 1;
    OnScreenRastPort = GetDC(MainWindowPtr);

    FillRect
    (   OnScreenRastPort,
        &therect,
        GetStockObject(GRAY_BRUSH)
    );

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}
EXPORT void rectfill_white(int leftx, int topy, int rightx, int bottomy)
{   RECT therect;
    HDC  OnScreenRastPort = NULL;

    assert(MainWindowPtr);

    therect.left   = leftx;
    therect.top    = topy;
    therect.right  = rightx + 1;
    therect.bottom = bottomy + 1;
    OnScreenRastPort = GetDC(MainWindowPtr);

    FillRect
    (   OnScreenRastPort,
        &therect,
        GetStockObject(WHITE_BRUSH)
    );

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}
EXPORT void rectfill_red(int leftx, int topy, int rightx, int bottomy)
{   RECT   therect;
    HDC    OnScreenRastPort = NULL;

    assert(MainWindowPtr);

    therect.left   = leftx;
    therect.top    = topy;
    therect.right  = rightx + 1;
    therect.bottom = bottomy + 1;
    OnScreenRastPort = GetDC(MainWindowPtr);

    FillRect
    (   OnScreenRastPort,
        &therect,
        RedBrush
    );

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void ami_draw(int firstx, int firsty, int secondx, int secondy, COLOUR colour)
{   HDC OnScreenRastPort = NULL;

    assert(MainWindowPtr);
    assert(colour == WHITE || colour == MEDIUMGREY || colour == BLACK);

    OnScreenRastPort = GetDC(MainWindowPtr);

    if (colour == WHITE)
    {   SelectObject(OnScreenRastPort, GetStockObject(WHITE_PEN));
    } elif (colour == MEDIUMGREY)
    {   SelectObject(OnScreenRastPort, GreyPen);
    } elif (colour == BLACK)
    {   SelectObject(OnScreenRastPort, GetStockObject(BLACK_PEN));
    }
    MoveToEx(OnScreenRastPort, firstx, firsty, NULL);
    if (firstx < secondx)
    {   secondx++;
    } elif (firstx > secondx)
    {   secondx--;
    }
    if (firsty < secondy)
    {   secondy++;
    } elif (firsty > secondy)
    {   secondy--;
    }
    LineTo(OnScreenRastPort, secondx, secondy);

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void le_loop(void)
{   FLAG            done = FALSE;
    int             mousex, mousey;
    MSG             Message;
    PAINTSTRUCT     ps;
    SCANCODE        scancode;
    POINT           ThePoint;
    RECT            therect;

    if (!MainWindowPtr)
    {   return;
    }

    globalabort = FALSE;
    while (a == LEVELEDIT && !globalabort)
    {   Sleep(0);
        while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
        {   switch (Message.message)
            {
            case WM_PAINT:
                BeginPaint(MainWindowPtr, &ps);
                redrawscreen();
                EndPaint(MainWindowPtr, &ps);
            acase WM_SYSKEYDOWN: // needed for F10
                // (Message.lParam & 0x40000000) means it is a repeated key
                scancode = PARAM2SCANCODE(Message.lParam);
                if (scancode == SCAN_F10 && !ctrl)
                {   le_handlekybd(scancode);
                } else
                {   TranslateMessage(&Message);
                    DispatchMessage(&Message);
                }
            acase WM_KEYDOWN:
                // (Message.lParam & 0x40000000) means it is a repeated key
                scancode = PARAM2SCANCODE(Message.lParam);
                switch (scancode)
                {
                case SCAN_LCTRL:
                case SCAN_RCTRL:
                    ctrl = TRUE;
                acase SCAN_LSHIFT:
                case SCAN_RSHIFT:
                    shift = TRUE;
                adefault:
                    effect(FXCLICK);
                    if (scancode == SCAN_N9 && Message.wParam == VK_PRIOR)
                    {   scancode = SCAN_PGUP;
                    } elif (scancode == SCAN_N3 && Message.wParam == VK_NEXT)
                    {   scancode = SCAN_PGDN;
                    }
                    if (ctrl)
                    {   le_handlectrlkybd(scancode);
                    } else
                    {   le_handlekybd(scancode);
                }   }
            acase WM_COMMAND:
                switch (LOWORD(Message.wParam))
                {
                case ID_FILE_NEW:
                    newfields();
                    say((STRPTR) GetCatalogStr(CatalogPtr, MSG_NEWDONE, "New done."), WHITE);
                acase ID_FILE_OPEN:
                    fileopen(FALSE);
                acase ID_FILE_REVERT:
                    fileopen(TRUE);
                acase ID_FILE_SAVE:
                    filesaveas(FALSE);
                acase ID_FILE_SAVEAS:
                    filesaveas(TRUE);
                acase ID_FILE_ICONIFY:
                    iconify();
                acase ID_FILE_EXITTITLE:
                    a = GAMEOVER;
                acase ID_FILE_EXITDOS:
                    if (verify())
                    {   cleanexit(EXIT_SUCCESS);
                    }
                acase ID_EDIT_CUT:
                    le_cut();
                acase ID_EDIT_COPY:
                    le_copy();
                acase ID_EDIT_PASTE:
                    le_paste();
                acase ID_EDIT_ERASE:
                    le_erase();
                acase ID_EDIT_DELETE:
                    le_delete();
                acase ID_EDIT_INSERT:
                    le_insert();
                acase ID_EDIT_APPEND:
                    le_append();
                acase ID_VIEW_PREV:
                    if (level == 0)
                    {   level = levels;
                    } else level--;
                    le_drawfield();
                acase ID_VIEW_NEXT:
                    if (level == levels)
                    {   level = 0;
                    } else level++;
                    le_drawfield();
                acase ID_HELP_CREATURES:
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CREATURES), MainWindowPtr, CreaturesDlgProc);
                acase ID_HELP_OBJECTS:
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_OBJECTS  ), MainWindowPtr, ObjectsDlgProc);
                acase ID_HELP_FRUITS:
                    DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_FRUITS   ), MainWindowPtr, FruitsDlgProc);
                acase ID_HELP_MANUAL:
                    ShellExecute(0, NULL, "WORMWARS.HTM", NULL, NULL, SW_SHOW);
                acase ID_HELP_UPDATE:
                    help_update();
                acase ID_HELP_ABOUT:
                    DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, AboutDlgProc);
                }
            acase WM_MOUSEMOVE:
                mousex = (int) (LOWORD(Message.lParam));
                mousey = (int) (HIWORD(Message.lParam));
                if (lmb)
                {   le_leftdown(mousex, mousey);
                }
                if (rmb)
                {   le_rightdown(mousex, mousey);
                }
            acase WM_LBUTTONDOWN:
            case WM_LBUTTONDBLCLK:
                mousex = (int) (LOWORD(Message.lParam));
                mousey = (int) (HIWORD(Message.lParam));
                lmb = TRUE;
                le_leftdown(mousex, mousey);
                if (mousex >= STARTXPIXEL && mousex <= ENDXPIXEL && mousey >= STARTYPIXEL && mousey <= ENDYPIXEL)
                {   confinepointer();
                }
            acase WM_LBUTTONUP:
                lmb = FALSE;
                releasepointer();
            acase WM_RBUTTONDOWN:
                rmb = TRUE;
                le_rightdown(mousex, mousey);
                if (mousex >= STARTXPIXEL && mousex <= ENDXPIXEL && mousey >= STARTYPIXEL && mousey <= ENDYPIXEL)
                {   confinepointer();
                }
            acase WM_RBUTTONUP:
                rmb = FALSE;
                releasepointer();
            acase WM_CLOSE:
                if (verify())
                {   cleanexit(EXIT_SUCCESS);
                }
            acase WM_KEYUP:
                switch (PARAM2SCANCODE(Message.lParam))
                {
                case SCAN_LCTRL:
                case SCAN_RCTRL:
                    ctrl = FALSE;
                acase SCAN_LSHIFT:
                case SCAN_RSHIFT:
                    shift = FALSE;
                }
            adefault:
                TranslateMessage(&Message);
                DispatchMessage(&Message);
}   }   }   }

EXPORT void ami_drawbox(int leftx, int topy, int width, int height)
{   assert(MainWindowPtr);

    /* WWWWW
       W   B
       W   B
       W   B
       BBBBB */

    ami_draw(leftx            , topy             , leftx            , topy + height - 1, WHITE); // left
    ami_draw(leftx + width - 1, topy + height - 1, leftx + width - 1, topy             , BLACK); // right
    ami_draw(leftx            , topy             , leftx + width - 1, topy             , WHITE); // top
    ami_draw(leftx            , topy + height - 1, leftx + width - 1, topy + height - 1, BLACK); // bottom
}

EXPORT void dot(void)
{   int x, xx, y, yy;
    HDC OnScreenRastPort = NULL;

    assert(MainWindowPtr);

    OnScreenRastPort = GetDC(MainWindowPtr);

    /* Squares are dotted as follows:

         012345678901
        0............
        1............
        2............
        3............
        4....WWW.....
        5....WWWB....
        6....WWWB....
        7.....BBB....
        8............
        9............
       10............
       11............ */

    xx = ((fex + WW_LEFTGAP) * SQUAREX) + STARTXPIXEL;
    yy = ((fey +  WW_TOPGAP) * SQUAREY) + STARTYPIXEL;

    for (x = 4; x <= 6; x++)
    {   for (y = 4; y <= 6; y++)
        {   DISCARD SetPixel(OnScreenRastPort, xx + x, yy + y, sticky ? RED : WHITE);
    }   }
    DISCARD SetPixel(OnScreenRastPort, xx + 7, yy + 5, BLACK);
    DISCARD SetPixel(OnScreenRastPort, xx + 7, yy + 6, BLACK);
    DISCARD SetPixel(OnScreenRastPort, xx + 5, yy + 7, BLACK);
    DISCARD SetPixel(OnScreenRastPort, xx + 6, yy + 7, BLACK);
    DISCARD SetPixel(OnScreenRastPort, xx + 7, yy + 7, BLACK);

    DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
    OnScreenRastPort = NULL;
}

EXPORT void updatemenu(void)
{   if (a != LEVELEDIT)
    {   return;
    }

    if (pathname[0])
    {   DISCARD EnableMenuItem(MenuPtr, ID_FILE_REVERT, MF_BYCOMMAND | MF_ENABLED);
    } else
    {   DISCARD EnableMenuItem(MenuPtr, ID_FILE_REVERT, MF_BYCOMMAND | MF_GRAYED);
    }

    if (levels > 1)
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_CUT   , MF_BYCOMMAND | MF_ENABLED);
        DISCARD EnableMenuItem(MenuPtr, ID_EDIT_DELETE, MF_BYCOMMAND | MF_ENABLED);
    } else
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_CUT   , MF_BYCOMMAND | MF_GRAYED);
        DISCARD EnableMenuItem(MenuPtr, ID_EDIT_DELETE, MF_BYCOMMAND | MF_GRAYED);
    }

    if (levels < MAXLEVELS)
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_INSERT, MF_BYCOMMAND | MF_ENABLED);
        DISCARD EnableMenuItem(MenuPtr, ID_EDIT_APPEND, MF_BYCOMMAND | MF_ENABLED);
    } else
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_INSERT, MF_BYCOMMAND | MF_GRAYED);
        DISCARD EnableMenuItem(MenuPtr, ID_EDIT_APPEND, MF_BYCOMMAND | MF_GRAYED);
    }

    if (clipboarded)
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_PASTE , MF_BYCOMMAND | MF_ENABLED);
    } else
    {   DISCARD EnableMenuItem(MenuPtr, ID_EDIT_PASTE , MF_BYCOMMAND | MF_GRAYED);
}   }

EXPORT FLAG verify(void)
{   TEXT tempstring[80 + 1];

    if (!levels_modified && !scores_modified)
    {   return TRUE;
    }
    if (autosave)
    {   // assert(levels_modified || scores_modified);
        if (pathname[0] == EOS)
        {   strcpy(pathname, DEFAULTSET);
        }
        filesaveas(FALSE);
        return TRUE;
    }

    if (levels_modified)
    {   sprintf
        (   tempstring,
            "%s %s",
            GetCatalogStr(CatalogPtr, MSG_MODIFIED,       "Level data has been modified."),
            GetCatalogStr(CatalogPtr, MSG_CONTINUE,       "Continue?")
        );
    } elif (scores_modified)
    {   sprintf
        (   tempstring,
            "%s %s",
            GetCatalogStr(CatalogPtr, MSG_MODIFIEDSCORES, "High scores have been modified."),
            GetCatalogStr(CatalogPtr, MSG_CONTINUE,       "Continue?")
        );
    } else
    {   return TRUE;
    }

    if (MessageBox
    (    CurrentWindowPtr,
         tempstring,
         "Worm Wars",
         MB_ICONQUESTION | MB_YESNO
    ) == IDYES)
    {   ctrl = shift = FALSE;
        return TRUE;
    }
    ctrl = shift = FALSE;
    return FALSE;
}

MODULE void le_handlectrlkybd(SCANCODE scancode)
{   assert(ctrl);

    switch (scancode)
    {
    case SCAN_F12:
        iconify();
    acase SCAN_A1:
    case SCAN_N1:
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_CREATURES), MainWindowPtr, CreaturesDlgProc);
    acase SCAN_A2:
    case SCAN_N2:
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_OBJECTS  ), MainWindowPtr, ObjectsDlgProc);
    acase SCAN_A3:
    case SCAN_N3:
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_FRUITS   ), MainWindowPtr, FruitsDlgProc);
    acase SCAN_SLASH:
        if (shift)
        {   DialogBox(InstancePtr, MAKEINTRESOURCE(IDD_ABOUT), MainWindowPtr, AboutDlgProc);
        }
    acase SCAN_A:
        filesaveas(TRUE);
    acase SCAN_C:
        le_copy();
    acase SCAN_E:
        le_erase();
    acase SCAN_M:
        ShellExecute(0, NULL, "WormWars.htm", NULL, NULL, SW_SHOW);
    acase SCAN_N:
        newfields();
        say((STRPTR) GetCatalogStr(CatalogPtr, MSG_NEWDONE, "New done."), WHITE);
    acase SCAN_O:
        fileopen(FALSE);
    acase SCAN_R:
        fileopen(TRUE);
    acase SCAN_S:
        filesaveas(FALSE);
    acase SCAN_V:
        le_paste();
    acase SCAN_X:
        le_cut();
    }

    ctrl = FALSE;
}

// For some fucked up reason, sscanf() with %f and atof() crash on
// Windows 7 and 8.1 (at least).
EXPORT double zatof(STRPTR inputstr)
{   int    decpoint = 0,
           i, j,
           length,
           start,
           temp;
    double result   = 0.0;

    if (inputstr[0] == '-')
    {   start = 1;
    } else
    {   start = 0;
    }

    length = strlen(inputstr);
    for (i = start; i < length; i++)
    {   if (inputstr[i] == '.')
        {   decpoint = i;
            break; // for speed
    }   }

    j = 1;
    for (i = decpoint - 1; i >= start; i--)
    {   temp = inputstr[i] - '0';
        result += (double) (temp * j);
        j *= 10;
    }

    j = 10;
    for (i = decpoint + 1; i < length; i++)
    {   if (inputstr[i] < '0' || inputstr[i] > '9') // eg. CR
        {   break;
        }
        temp = inputstr[i] - '0';
        result += (double) (((double) temp) / j);
        j *= 10;
    }

    if (start == 1)
    {   result = -result;
    }

    return result;
}

EXPORT void openurl(STRPTR command)
{   DISCARD ShellExecute(0, NULL, command, NULL, NULL, SW_SHOW);
}

EXPORT void raw_draw(int x, int y, UWORD what)
{   HDC     CompatibleRastPort;
    HBITMAP OldBitmapPtr;
    FLAG    release;

    if (OnScreenRastPort)
    {   release = FALSE;
    } else
    {   release = TRUE;
        OnScreenRastPort = GetDC(MainWindowPtr);
    }

    CompatibleRastPort = CreateCompatibleDC(OnScreenRastPort);
    OldBitmapPtr = (HBITMAP) SelectObject(CompatibleRastPort, SquaresBitmap);
    BitBlt(OnScreenRastPort, x, y, SQUAREX, SQUAREY, CompatibleRastPort, (what % 16) * SQUAREX, (what / 16) * SQUAREY, SRCCOPY);
    SelectObject(CompatibleRastPort, OldBitmapPtr);
    DeleteDC(CompatibleRastPort);
    if (release)
    {   DISCARD ReleaseDC(MainWindowPtr, OnScreenRastPort);
        OnScreenRastPort = NULL;
}   }

EXPORT void intro_draw(int x, int y, UWORD what)
{   raw_draw
    (   STARTXPIXEL + (((coordfieldx / 2) + x) * SQUAREX),
        yoffset - (SQUAREY * 11) + (SQUAREY * y),
        what
    );
}

MODULE void setcodepage(void)
{   UINT cp;
    TEXT tempstring[80 + 1];

    if (consoleopen)
    {   // This call is always returning 1 under WinXP (regardless of user's Control Panel settings, ie. regardless of whether it was successful or not)
        DISCARD SetConsoleOutputCP(langs[language].codepage);
    }

    cp = GetACP();
    if (cp != (UINT) langs[language].codepage)
    {   /* None of these work here (on Win8.1) :-( but they do work when typed at the command line.
        WinExec("%systemroot%\\system32\\control.exe /name Microsoft.RegionalAndLanguageOptions /page /p:\"administrative\"", SW_NORMAL);
        WinExec("%systemroot%\\system32\\control.exe international", SW_NORMAL);
        openurl("control.exe international"); */

        sprintf(tempstring, "Incorrect code page (%d expected, %d found)!\n", langs[language].codepage, cp); // localize this!
        say(tempstring, RED);
}   }

MODULE void changelanguage(void)
{   if (language != LANGUAGE_ENG)
    {   load_catalog();
    }
    setcodepage();
    enginesetup2();
    retranslate();
}

EXPORT void busypointer(void)
{   SetCursor(hBusy);
}
EXPORT void normalpointer(void)
{   SetCursor(hPointer);
}

MODULE void confinepointer(void)
{   POINT ThePoint;
    RECT  therect;

    ThePoint.x = 0;
    ThePoint.y = 0;
    DISCARD ClientToScreen(MainWindowPtr, &ThePoint);
    therect.left   = ThePoint.x + STARTXPIXEL;
    therect.right  = ThePoint.x + STARTXPIXEL + ((MINFIELDX + 1) * SQUAREX);
    therect.top    = ThePoint.y + STARTYPIXEL;
    therect.bottom = ThePoint.y + STARTYPIXEL + ((MINFIELDY + 1) * SQUAREY);
    DISCARD ClipCursor(&therect);
}
EXPORT void releasepointer(void)
{   DISCARD ClipCursor(NULL);
}

MODULE void drawitem(LPDRAWITEMSTRUCT lpDIS)
{   int    centrex, centrey,
           totalx,
           extraxy;
    ULONG  stateflag;
    RECT   localrect = lpDIS->rcItem;
    SIZE   localsize;
    HICON  localicon;
    HDC    hdc       = lpDIS->hDC;
    STRPTR stringptr1, stringptr2;

    switch (lpDIS->CtlID)
    {
    // No ampersand for stringptr1! otherwise text size is calculated wrongly
    case IDC_LOAD:
        stringptr1 = LLL(MSG_OPENLSET      , "Load levelset..." );
        stringptr2 = LLL(MSG_OPENLSET2     , "&Load levelset...");
        localicon = icons[0];
    acase IDC_SAVE:
        stringptr1 = LLL(MSG_SAVELSET      , "Save levelset..." );
        stringptr2 = LLL(MSG_SAVELSET2     , "&Save levelset...");
        localicon = icons[1];
    acase IDC_MANUAL:
        stringptr1 = LLL(MSG_MANUAL2       , "Manual..."        );
        stringptr2 = LLL(MSG_MANUAL        , "Ma&nual..."       );
        localicon = icons[2];
    acase IDC_ABOUT:
        stringptr1 = LLL(MSG_ABOUT2        , "About..."         );
        stringptr2 = LLL(MSG_ABOUT         , "&About..."        );
        localicon = icons[3];
    acase IDC_CREATURES:
        stringptr1 = LLL(MSG_MENU_CREATURES, "Creatures..."     );
        stringptr2 = LLL(MSG_CREATURES     , "&Creatures..."    );
        localicon = icons[4];
    acase IDC_OBJECTS:
        stringptr1 = LLL(MSG_MENU_OBJECTS  , "Objects..."       );
        stringptr2 = LLL(MSG_OBJECTS       , "&Objects..."      );
        localicon = icons[5];
    acase IDC_FRUITS:
        stringptr1 = LLL(MSG_MENU_FRUITS   , "Fruits..."        );
        stringptr2 = LLL(MSG_FRUITS        , "&Fruits..."       );
        localicon = icons[6];
    acase IDC_START:
        stringptr1 = LLL(MSG_STARTGAME3    , "Start Game"       );
        stringptr2 = LLL(MSG_STARTGAME2    , "Start &Game"      );
        localicon = icons[7];
    acase IDC_LEVELEDITOR:
        stringptr1 = LLL(MSG_LEVELEDITOR   , "Level Editor"     );
        stringptr2 = LLL(MSG_LEVELEDITOR2  , "Level &Editor"    );
        localicon = icons[8];
    acase IDC_UPDATE:
        stringptr1 =
        stringptr2 = LLL(MSG_MENU_UPDATE   , "Check for updates...");
        localicon = icons[9];
    acase IDC_CLEARHISCORES:
        stringptr1 =
        stringptr2 = LLL(MSG_MENU_CLEARHS  , "Clear high scores");
        localicon = icons[10];
    acase IDOK:
        stringptr1 = LLL(MSG_OK2           , "OK"               );
        stringptr2 = LLL(MSG_OK            , "&OK"              );
        localicon = icons[11];
    }

    if (lpDIS->itemState & ODS_SELECTED)
        {   DrawFrameControl(hdc, &localrect, DFC_BUTTON, DFCS_BUTTONPUSH | DFCS_PUSHED);
        extraxy = 1;
    } else
    {   DrawFrameControl(hdc, &localrect, DFC_BUTTON, DFCS_BUTTONPUSH);
        extraxy = 0;
    }
    if (lpDIS->itemState & ODS_DISABLED)
    {   stateflag = DSS_DISABLED;
    } else
    {   stateflag = DSS_NORMAL;
        }

    centrex = (localrect.right  - localrect.left + 1) / 2;
    centrey = (localrect.bottom - localrect.top  + 1) / 2;

    DISCARD GetTextExtentPoint32
    (   hdc,                // handle to device context
        stringptr1,         // pointer to text string
        strlen(stringptr1), // number of characters in string
        &localsize          // pointer to structure for string size
    );
    totalx = localsize.cx + 16 + 4;

    DrawState
    (   hdc,
        NULL,
        NULL,
        (LPARAM) localicon,
        (WPARAM) NULL,
        centrex - (totalx / 2) + extraxy,
        centrey - (16     / 2) + extraxy,
        16, 16,
        DST_ICON | stateflag
    );

    DrawState
    (   hdc,
        NULL,
        NULL,
        (LPARAM) stringptr2,
        (WPARAM) NULL,
        centrex - (totalx       / 2) + 16 + 4 + extraxy,
        centrey - (localsize.cy / 2)      - 1 + extraxy,
        localsize.cx,
        localsize.cy,
        DST_PREFIXTEXT | stateflag
    );

    if (!(lpDIS->itemState & ODS_DISABLED) && (lpDIS->itemState & ODS_FOCUS))
        {   localrect.left   += 3;
        localrect.right  -= 3;
        localrect.top    += 3;
        localrect.bottom -= 3;
        DrawFocusRect(hdc, &localrect);
}   }

MODULE void do_systemmenu(HWND hwnd)
{   HMENU SystemMenuPtr;

    SystemMenuPtr = GetSystemMenu(hwnd, FALSE);
    if (!SystemMenuPtr)
    {   return;
    }

    DISCARD InsertMenu(SystemMenuPtr,  6, MF_BYPOSITION | MF_STRING   , ID_HELP_MANUAL , LLL(MSG_MANUAL,       "Ma&nual..."          ));
    DISCARD SetMenuItemBitmaps
    (   SystemMenuPtr,             // handle of menu
        mib[MIB_MANUAL].menuitem,  // menu item to receive new bitmaps
        MF_BYCOMMAND,              // menu item flags
        mib[MIB_MANUAL].bitmap,    // handle of unchecked bitmap
        NULL                       // handle of checked bitmap
    );
    DISCARD InsertMenu(SystemMenuPtr,  7, MF_BYPOSITION | MF_STRING   , ID_HELP_UPDATE , LLL(MSG_MENU_UPDATE,  "Check for updates..."));
    DISCARD SetMenuItemBitmaps
    (   SystemMenuPtr,             // handle of menu
        mib[MIB_UPDATES].menuitem, // menu item to receive new bitmaps
        MF_BYCOMMAND,              // menu item flags
        mib[MIB_UPDATES].bitmap,   // handle of unchecked bitmap
        NULL                       // handle of checked bitmap
    );
    DISCARD InsertMenu(SystemMenuPtr,  8, MF_BYPOSITION | MF_STRING   , ID_HELP_ABOUT  , LLL(MSG_MENU_ABOUTWW, "About Worm Wars..."  ));
    DISCARD SetMenuItemBitmaps
    (   SystemMenuPtr,             // handle of menu
        mib[MIB_ABOUT].menuitem,   // menu item to receive new bitmaps
        MF_BYCOMMAND,              // menu item flags
        mib[MIB_ABOUT].bitmap,     // handle of unchecked bitmap
        NULL                       // handle of checked bitmap
    );
    DISCARD InsertMenu(SystemMenuPtr,  9, MF_BYPOSITION | MF_SEPARATOR, 0              , 0);
    DISCARD InsertMenu(SystemMenuPtr, 10, MF_BYPOSITION | MF_STRING   , ID_FILE_ICONIFY, "&Iconify\tCtrl+F12"); // should localize this
    DISCARD SetMenuItemBitmaps
    (   SystemMenuPtr,             // handle of menu
        mib[MIB_ICONIFY].menuitem, // menu item to receive new bitmaps
        MF_BYCOMMAND,              // menu item flags
        mib[MIB_ICONIFY].bitmap,   // handle of unchecked bitmap
        NULL                       // handle of checked bitmap
    );
}
